from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from flow.record.fieldtypes import path

from dissect.target.plugins.apps.browser import iexplore
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target

# NOTE: Missing cookies, extensions and passwords tests.


@pytest.fixture
def target_iexplore(target_win_users: Target, fs_win: VirtualFilesystem) -> Target:
    base_path = "Users\\John\\AppData\\Local\\Microsoft\\Windows\\WebCache"
    files = [
        ("WebCacheV01.dat", "_data/plugins/apps/browser/iexplore/WebCacheV01.dat.gz", "gzip"),
    ]

    for filename, test_path, compression in files:
        fs_win.map_file(f"{base_path}\\{filename}", absolute_path(test_path), compression=compression)

    target_win_users.add_plugin(iexplore.InternetExplorerPlugin)

    return target_win_users


def test_iexplore_history(target_iexplore: Target) -> None:
    records = list(target_iexplore.iexplore.history())
    assert len(records) == 41

    # Test namespaced plugin
    records = list(target_iexplore.browser.history())
    assert len(records) == 41


def test_iexplore_downloads(target_iexplore: Target) -> None:
    records = list(target_iexplore.iexplore.downloads())
    assert len(records) == 1
    assert records[0].path == path.from_windows("C:\\Users\\John\\Downloads\\archlinux-2023.02.01-x86_64.iso")
    assert records[0].url == "https://mirror.cj2.nl/archlinux/iso/2023.02.01/archlinux-2023.02.01-x86_64.iso"



from __future__ import annotations

import csv
import os
from enum import Enum
from typing import TYPE_CHECKING, NamedTuple

from dissect.target.helpers.logging import get_logger

if TYPE_CHECKING:
    from pathlib import Path


log = get_logger(__name__)


class KeyType(Enum):
    """Valid key types."""

    RAW = "raw"
    PASSPHRASE = "passphrase"
    RECOVERY_KEY = "recovery_key"
    FILE = "file"


class Key(NamedTuple):
    key_type: KeyType
    value: str | bytes
    provider: str | None = None
    identifier: str | None = None
    is_wildcard: bool = False


KEYCHAIN: list[Key] = []


def register_key(
    key_type: KeyType, value: str, identifier: str | None = None, provider: str | None = None, is_wildcard: bool = False
) -> None:
    if key_type == KeyType.RAW:
        try:
            value = bytes.fromhex(value)
        except ValueError:
            log.debug("Failed to decode raw key as hex, ignoring: %s", value)
            return

    if key_type in (KeyType.RECOVERY_KEY, KeyType.FILE):
        value = value.strip("\"'")

    key = Key(key_type, value, provider, identifier, is_wildcard)
    KEYCHAIN.append(key)
    log.info("Registered key %s", key)


def get_all_keys() -> list[Key]:
    return KEYCHAIN[:]


def get_keys_for_provider(provider: str) -> list[Key]:
    return [key for key in KEYCHAIN if key.provider and key.provider.lower() == provider.lower()]


def get_keys_without_provider() -> list[Key]:
    return [key for key in KEYCHAIN if not key.provider]


def parse_key_type(key_type_name: str) -> KeyType:
    for key_type in KeyType:
        if key_type.value == key_type_name:
            return key_type
    raise ValueError("No KeyType enum values that match %s", key_type_name)


def register_wildcard_value(value: str) -> None:
    for key_type in KeyType:
        register_key(key_type, value, is_wildcard=True)


def register_keychain_file(keychain_path: Path) -> None:
    """Register all keys from provided keychain file.

    The keychain file is a CSV file in "excel" dialect with the columns:

        provider,key_type,identifier,value

    Values in columns ``key_type`` and ``value`` are required, ``provider`` and ``identifier`` are optional.
    Example of the CSV-serialised data:

        bitlocker,recovery_key,,312268-409816-583517-486695-627121-599511-664389-145640
        bitlocker,passphrase,,Password1234
        ,passphrase,,AnotherTestPassword
        bitlocker,passphrase,b6ad258a-2725-4a42-93c6-844478bf7a90,Password1234
        bitlocker,passphrase,,"Password with comma, space and quotes ' and "" inside"
    """
    if not keychain_path.exists():
        raise ValueError("Provided keychain file %s does not exists", keychain_path)

    with keychain_path.open(mode="r") as fh:
        rows = csv.reader(fh)
        for row in rows:
            # Skip comment rows (starting with #)
            if row[0].startswith("#"):
                continue
            if len(row) != 4:
                log.warning("Expecting 4 values per row in a keychain CSV file, found %d: %s. Skipping", len(row), row)
                continue

            provider, key_type_name, identifier, value = row
            try:
                key_type = parse_key_type(key_type_name)
            except ValueError:
                log.warning("Unrecognised key type %s", key_type_name)
                continue

            if not value:
                log.warning("No value provided in row %s", row)
                continue

            identifier = identifier if identifier else None
            provider = provider if provider else None

            register_key(
                key_type=key_type,
                value=value,
                identifier=identifier,
                provider=provider,
            )


# Add DISSECT_KEYCHAIN_VALUE from environment to KEYCHAIN on module load
if _env_value := os.getenv("DISSECT_KEYCHAIN_VALUE"):
    register_wildcard_value(_env_value)



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.util.ts import wintimestamp

from dissect.target.helpers.regutil import VirtualHive, VirtualKey, VirtualValue
from dissect.target.plugins.os.windows.cam import CamPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_cam_history(target_win_users: Target, hive_hklm: VirtualHive, fs_win: VirtualFilesystem) -> None:
    fs_win.map_file(
        "ProgramData/Microsoft/Windows/CapabilityAccessManager/CapabilityAccessManager.db",
        absolute_path("_data/plugins/os/windows/cam/CapabilityAccessManager.db"),
    )

    hklm_cam_key = VirtualKey(
        hive_hklm,
        "Software\\Microsoft\\Windows\\CurrentVersion\\CapabilityAccessManager\\CapabilityUsageHistory",
    )

    database_root_value = "C:\\ProgramData\\Microsoft\\Windows\\CapabilityAccessManager"
    hklm_cam_key.add_value("DatabaseRoot", VirtualValue(hive_hklm, "DatabaseRoot", database_root_value))

    hive_hklm.map_key(hklm_cam_key.path, hklm_cam_key)

    target_win_users.add_plugin(CamPlugin)
    results = list(target_win_users.cam())

    assert len(results) == 3

    # Record windows/cam/usagehistory - NonPackagedUsageHistory
    assert results[0].last_used_time_stop == wintimestamp(133885054906926593)
    assert results[0].last_used_time_start == wintimestamp(133885044556858623)
    assert results[0].duration == 1035
    assert results[0].package_type == "NonPackagedUsageHistory"
    assert results[0].capability == "microphone"
    assert results[0].file_id == "0000f0bfca16305374262ad1919c258deba64fc25006"
    assert results[0].file_id_hash.sha1 == "f0bfca16305374262ad1919c258deba64fc25006"
    assert results[0].access_blocked == "0"
    assert results[0].program_id == "00001495bc9d3674a3db3bbed7de63ca0f920000ffff"
    assert results[0].package_family_name is None
    assert results[0].access_guid is None
    assert results[0].label == "2"
    assert results[0].app_name == "Microsoft Edge"
    assert results[0].binary_full_path == "C:\\Program Files (x86)\\Microsoft\\Edge\\Application\\msedge.exe"
    assert results[0].service_name is None
    assert results[0].username is None

    # Record windows/cam/usagehistory - NonPackagedUsageHistory
    assert results[1].last_used_time_stop == wintimestamp(133884984261774039)
    assert results[1].last_used_time_start == wintimestamp(133884984204276055)
    assert results[1].duration == 5
    assert results[1].package_type == "PackagedUsageHistory"
    assert results[1].capability == "microphone"
    assert results[1].file_id is None
    assert results[1].file_id_hash.sha1 is None
    assert results[1].access_blocked == "0"
    assert results[1].program_id is None
    assert results[1].package_family_name == "Microsoft.WindowsSoundRecorder_8wekyb3d8bbwe"
    assert results[1].app_name == "Sound Recorder"
    assert results[1].access_guid is None
    assert results[1].label == "2"
    assert results[1].app_name == "Sound Recorder"
    assert results[1].binary_full_path is None
    assert results[1].service_name is None
    assert results[1].username is None

    # Record windows/cam/identityrelationshiphistory - NonPackagedIdentityRelationship
    assert results[2].last_observed_time == wintimestamp(133892048101847846)
    assert results[2].file_id == "0000565c728f1a97551d85db2e788c69e0d8a18ea777"
    assert results[2].file_id_hash.sha1 == "565c728f1a97551d85db2e788c69e0d8a18ea777"
    assert results[2].program_id == "0000d44ffb9f3c146e6da82376ed56489aff0000ffff"
    assert results[2].binary_full_path == "C:\\Program Files (x86)\\Microsoft\\Edge\\Application\\msedge.exe"

    # Record windows/cam/globalprompthistory - NonPackagedGlobalPromptHistory
    # No test data for this record type.


def test_cam_registry(target_win_users: Target, hive_hku: VirtualHive, hive_hklm: VirtualHive) -> None:
    hku_cam_key = VirtualKey(
        hive_hku,
        "Software\\Microsoft\\Windows\\CurrentVersion\\CapabilityAccessManager\\ConsentStore\\webcam",
    )

    hklm_cam_key = VirtualKey(
        hive_hklm,
        "Software\\Microsoft\\Windows\\CurrentVersion\\CapabilityAccessManager\\ConsentStore\\microphone",
    )

    microsoft_camera_key = VirtualKey(
        hive_hku, f"{hive_hku}\\{hku_cam_key.path}\\Microsoft.WindowsCamera_8wekyb3d8bbwe"
    )
    microsoft_camera_key.add_value("LastUsedTimeStart", VirtualValue(hive_hku, "LastUsedTimeStart", 133784711366495858))
    microsoft_camera_key.add_value("LastUsedTimeStop", VirtualValue(hive_hku, "LastUsedTimeStop", 133784711515887950))

    firefox_key = VirtualKey(
        hive_hku, f"{hive_hku}\\{hku_cam_key.path}\\NonPackaged\\C:#Program Files#Mozilla Firefox#firefox.exe"
    )
    firefox_key.add_value("LastUsedTimeStart", VirtualValue(hive_hku, "LastUsedTimeStart", 133788466426086184))
    firefox_key.add_value("LastUsedTimeStop", VirtualValue(hive_hku, "LastUsedTimeStop", 133788466774490036))

    nonpackaged_key = VirtualKey(hive_hku, f"{hive_hku}\\{hku_cam_key.path}\\NonPackaged")
    nonpackaged_key.add_subkey("C:#Program Files#Mozilla Firefox#firefox.exe", firefox_key)

    python_key = VirtualKey(
        hive_hklm, f"{hive_hklm}\\{hklm_cam_key.path}\\PythonSoftwareFoundation.Python.3.11_qbz5n2kfra8p0"
    )
    python_key.add_value("LastUsedTimeStart", VirtualValue(hive_hklm, "LastUsedTimeStart", 133788466426086163))
    python_key.add_value("LastUsedTimeStop", VirtualValue(hive_hklm, "LastUsedTimeStop", 133788466774490044))

    hku_cam_key.add_subkey(microsoft_camera_key.name, microsoft_camera_key)
    hku_cam_key.add_subkey(nonpackaged_key.name, nonpackaged_key)
    hklm_cam_key.add_subkey(python_key.name, python_key)

    hive_hku.map_key(hku_cam_key.path, hku_cam_key)
    hive_hklm.map_key(hklm_cam_key.path, hklm_cam_key)

    target_win_users.add_plugin(CamPlugin)
    results = list(target_win_users.cam())

    assert len(results) == 3
    assert results[0].device == "webcam"
    assert results[0].app_name == "Microsoft.WindowsCamera_8wekyb3d8bbwe"
    assert results[0].path is None
    assert results[0].last_started == wintimestamp(133784711366495858)
    assert results[0].last_stopped == wintimestamp(133784711515887950)
    assert results[1].device == "webcam"
    assert results[1].app_name == "C:#Program Files#Mozilla Firefox#firefox.exe"
    assert results[1].path == "C:\\Program Files\\Mozilla Firefox\\firefox.exe"
    assert results[1].last_started == wintimestamp(133788466426086184)
    assert results[1].last_stopped == wintimestamp(133788466774490036)
    assert results[2].device == "microphone"
    assert results[2].app_name == "PythonSoftwareFoundation.Python.3.11_qbz5n2kfra8p0"
    assert results[2].path is None
    assert results[2].last_started == wintimestamp(133788466426086163)
    assert results[2].last_stopped == wintimestamp(133788466774490044)



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING

from dissect.target.plugins.apps.webhosting.cpanel import CPanelPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_cpanel_plugin(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    data_file = absolute_path("_data/plugins/apps/webhosting/cpanel/lastlogin")
    fs_unix.map_file("/home/user/.lastlogin", data_file)

    fs_unix.makedirs("/usr/local/cpanel/logs")

    target_unix_users.add_plugin(CPanelPlugin)

    results = list(target_unix_users.cpanel.lastlogin())

    record = results[0]

    assert len(results) == 6
    assert record.ts == datetime(2023, 6, 27, 13, 22, 13, tzinfo=timezone.utc)
    assert record.user == "user"
    assert record.remote_ip == "8.8.8.8"



from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING

from dissect.target.plugins.os.unix.linux.services import ServicesPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_services(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    systemd_service_1 = absolute_path("_data/plugins/os/unix/services/systemd.service")
    systemd_service_2 = absolute_path("_data/plugins/os/unix/services/systemd2.service")
    initd_service_1 = absolute_path("_data/plugins/os/unix/services/initd.sh")
    fs_unix.map_file("/lib/systemd/system/example.service", systemd_service_1)
    fs_unix.symlink("/dev/null", "/lib/systemd/system/broken_sym_example.service")
    fs_unix.map_file("/usr/lib/systemd/system/example2.service", systemd_service_2)
    fs_unix.map_file("/etc/init.d/example", initd_service_1)
    target_unix_users.add_plugin(ServicesPlugin)

    results = list(target_unix_users.services())
    assert len(results) == 4

    result_0 = results[0]
    assert result_0.name == "example.service"
    expected_config = {
        "Unit_Description": "an example systemd service",
        "Unit_After": "foobar.service",
        "Unit_Requires": "foo.service bar.service",
        "Service_Type": "simple",
        "Service_ExecStart": "/usr/sbin/simple-command --key value",
        "Service_SyslogIdentifier": None,
        "Service_TimeoutStopSec": "5",
        "Install_WantedBy": "multi-user.target",
        "Install_Alias": "example.service",
    }
    for key, value in expected_config.items():
        assert getattr(result_0, key) == value

    assert str(result_0.source) == "/lib/systemd/system/example.service"

    assert results[1].name == "broken_sym_example.service"
    assert str(results[1].source) == "/lib/systemd/system/broken_sym_example.service"

    result_2 = results[2]
    assert result_2.name == "example2.service"
    expected_config_2 = {
        "Unit_Description": "an example systemd service",
        "Unit_After": "foobar.service",
        "Unit_Requires": "foo.service bar.service",
        "Service_Type": "simple",
        "Service_ExecStart": (
            "/bin/bash -c 'exec /usr/bin/example param1 --param2=value2 -P3value3 -param4 value4; exit 0'"
        ),
        "Service_SyslogIdentifier": "example-service",
        "Service_TimeoutStopSec": "5",
        "Install_WantedBy": "multi-user.target",
        "Install_Alias": "example.service",
    }

    for key, value in expected_config_2.items():
        assert getattr(result_2, key) == value
    assert str(result_2.source) == "/usr/lib/systemd/system/example2.service"

    assert results[3].name == "example"
    assert str(results[3].source) == "/etc/init.d/example"


def test_services_systemd_drop_folder(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if we correctly parse systemd ``.service`` and drop folder ``.conf`` files. Data based on Ubuntu 22.04 LTS default files."""  # noqa: E501

    fs_unix.map_file(
        "/lib/systemd/system/systemd-localed.service",
        str(absolute_path("_data/plugins/os/unix/linux/services/systemd-localed.service")),
    )
    fs_unix.map_file(
        "/lib/systemd/system/systemd-localed.service.d/locale-gen.conf",
        str(absolute_path("_data/plugins/os/unix/linux/services/locale-gen.conf")),
    )
    fs_unix.map_file_fh(
        "/lib/systemd/system/systemd-localed.service.d/test.conf",
        BytesIO(b"[Unit]\nDescription=Another Description\n"),
    )

    target_unix.add_plugin(ServicesPlugin)
    results = list(target_unix.services())

    # Make sure we convert duplicate values to a list
    assert results[0].Unit_Description == ["Locale Service", "Another Description"]

    # Make sure we convert multiple entries to a list
    assert results[0].Unit_Documentation == [
        "man:systemd-localed.service(8)",
        "man:locale.conf(5)",
        "man:vconsole.conf(5)",
        "man:org.freedesktop.locale1(5)",
    ]

    # Make sure we map empty values to None
    assert results[0].Service_CapabilityBoundingSet is None

    # Make sure we merge .service entries and drop file entries into a single list
    assert results[0].Service_ReadWritePaths == ["/etc", "/usr/lib/locale", "/usr/lib/locale/"]

    # We do not have visibility in the parsed drop files since that is handled internally by the drop file configparser
    assert results[0].source == "/lib/systemd/system/systemd-localed.service"



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import ChildTargetRecord
from dissect.target.plugin import ChildTargetPlugin

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from dissect.target import Target


def find_vms(path: Path) -> Iterator[tuple[str, Path, Path]]:
    """Find the Lima VMs from Colima and yield the name, Colima configuration path and Lima VM path.

    References:
        - https://github.com/abiosoft/colima/blob/5ddf1e0dc67772f6e28f84c7c7b32f2343ad4bfb/config/profile.go#L19-L39
    """
    for config_file in path.glob("*/colima.yaml"):
        profile = config_file.parent.name
        if profile == "default":
            lima_id = "colima"
        else:
            profile = profile.removeprefix("colima-")
            lima_id = f"colima-{profile}"

        if (lima_path := path.joinpath("_lima", lima_id)).exists():
            yield profile, config_file, lima_path


class ColimaChildTargetPlugin(ChildTargetPlugin):
    """Child target plugin that yields Colima VMs.

    Colima is a container runtime for macOS and Linux.

    References:
        - https://github.com/abiosoft/colima
    """

    __type__ = "colima"

    def __init__(self, target: Target):
        super().__init__(target)
        self.paths = []
        for user in self.target.user_details.all_with_home():
            # check .colima folder in ~/
            if (path := user.home_path.joinpath(".colima")).exists():
                self.paths.append(path)
            # check .colima folder in ~/.config/
            if (path := user.home_path.joinpath(".config", "colima")).exists():
                self.paths.append(path)

    def check_compatible(self) -> None:
        if not self.paths:
            raise UnsupportedPluginError("No Colima configurations found")

    def list_children(self) -> Iterator[ChildTargetRecord]:
        for path in self.paths:
            for name, _, lima_path in find_vms(path):
                yield ChildTargetRecord(
                    type=self.__type__,
                    name=name,
                    path=lima_path.joinpath("diffdisk"),
                    _target=self.target,
                )



from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING
from unittest.mock import Mock

from flow.record.fieldtypes import posix_path

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.plugin import OperatingSystem
from dissect.target.plugins.os.unix.bsd.darwin.macos._os import MacOSPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.target import Target


def test_macos_os(target_macos_users: Target, fs_macos: VirtualFilesystem) -> None:
    """Test if we detect a macOS target correctly."""

    target_macos_users.add_plugin(MacOSPlugin)

    interface = absolute_path("_data/plugins/os/unix/bsd/darwin/macos/_os/en0.plist")
    fs_macos.map_file("/private/var/db/dhcpclient/leases/en0.plist", interface)

    users = list(target_macos_users.users())
    ips = target_macos_users.ips
    ips.sort()

    dissect_user = users[0]
    test_user = users[1]

    assert target_macos_users.os == OperatingSystem.MACOS
    assert target_macos_users.hostname == "dummys-Mac"
    assert target_macos_users.version == "macOS 11.7.5 (20G1225)"

    assert len(users) == 2
    assert len(ips) == 2

    assert dissect_user._desc.name == "macos/user"
    assert dissect_user.name == "_dissect"
    assert dissect_user.passwd == "*"
    assert dissect_user.home == posix_path("/Users/dissect")
    assert dissect_user.shell == "/usr/bin/false"
    assert dissect_user.source == "/var/db/dslocal/nodes/Default/users/_dissect.plist"

    assert test_user.home is None

    assert ips == ["10.42.43.63", "10.42.43.64"]


def test_apfs_mounts(target_bare: Target) -> None:
    """Test that macOS firmlinks are correctly mounted."""
    mock_system_volume = VirtualFilesystem()
    mock_system_volume.__type__ = "apfs"
    mock_system_volume.apfs = Mock()
    mock_system_volume.apfs.role = Mock()
    mock_system_volume.apfs.role.name = "SYSTEM"

    mock_system_volume.makedirs("/Library")
    mock_system_volume.makedirs("/Applications")
    mock_system_volume.map_file_fh("/usr/share/firmlinks", BytesIO(b"/Applications\tApplications\n/Library\tLibrary\n"))

    mock_data_volume = VirtualFilesystem()
    mock_data_volume.__type__ = "apfs"
    mock_data_volume.apfs = Mock()
    mock_data_volume.apfs.role = Mock()
    mock_data_volume.apfs.role.name = "DATA"
    mock_data_volume.makedirs("/Library/something")
    mock_data_volume.makedirs("/Applications/Some.app")

    target_bare.filesystems.add(mock_system_volume)
    target_bare.filesystems.add(mock_data_volume)

    assert MacOSPlugin.detect(target_bare) is mock_system_volume
    target_bare._os_plugin = MacOSPlugin
    target_bare.apply()

    assert target_bare.fs.exists("/Applications")
    assert target_bare.fs.exists("/Library")
    assert target_bare.fs.exists("/Applications/Some.app")
    assert target_bare.fs.exists("/Library/something")



from __future__ import annotations

from typing import TYPE_CHECKING

from flow.record.fieldtypes import datetime

from dissect.target.plugins.os.windows.log.schedlgu import SchedLgUPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_shedlgu(target_win: Target, fs_win: VirtualFilesystem) -> None:
    shedlgu_file = absolute_path("_data/plugins/os/windows/log/schedlgu/schedlgu.txt")
    fs_win.map_file("Windows/SchedLgU.Txt", shedlgu_file)

    target_win.add_plugin(SchedLgUPlugin)

    records = list(target_win.schedlgu())
    task_scheduler_started_event = records[0]
    task_scheduler_version_event = records[1]
    task_scheduler_exited_event = records[2]
    job_task_event = records[58]

    assert task_scheduler_started_event.ts == datetime("2006-11-02 07:35:17+00:00")
    assert task_scheduler_started_event.job == "Task Scheduler Service"
    assert task_scheduler_started_event.status == "Started"

    assert task_scheduler_version_event.job == "Task Scheduler Service"
    assert task_scheduler_version_event.version == "6.0.6000.16386 (vista_rtm.061101-2205)"

    assert task_scheduler_exited_event.job == "Task Scheduler Service"
    assert task_scheduler_exited_event.ts == datetime("2006-11-02 07:55:10+00:00")
    assert task_scheduler_exited_event.status == "Exited"

    assert job_task_event.ts == datetime("2003-09-14 13:01:00+00:00")
    assert job_task_event.job == "Symantec NetDetect.job"
    assert job_task_event.command == "NDETECT.EXE"
    assert job_task_event.status == "Finished"
    assert job_task_event.exit_code == 65



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import patch

import pytest

from dissect.target.loader import open as loader_open
from dissect.target.loaders.tar import TarLoader
from dissect.target.loaders.uac import UacLoader, UacTarSubloader, UacZipSubLoader
from dissect.target.loaders.zip import ZipLoader
from dissect.target.target import Target
from tests._utils import absolute_path, mkdirs

if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path

    from pytest_benchmark.fixture import BenchmarkFixture

    from dissect.target.loader import Loader


@pytest.fixture
def mock_uac_dir(tmp_path: Path) -> Path:
    root = tmp_path
    mkdirs(root / "[root]", ["etc", "var"])
    (root / "uac.log").write_bytes(b"")
    (root / "[root]" / "etc" / "passwd").write_bytes(b"root:x:0:0:root:/root:/bin/bash\n")
    return tmp_path


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
@pytest.mark.parametrize(
    ("path", "loader"),
    [
        ("_data/loaders/uac/uac-2e44ea6da71d-linux-20250717143111.tar.gz", TarLoader),
        ("_data/loaders/uac/uac-2e44ea6da71d-linux-20250717143112.tar.gz", TarLoader),
        ("_data/loaders/uac/uac-2e44ea6da71d-linux-20250717143106.zip", ZipLoader),
        ("mock_uac_dir", UacLoader),
    ],
)
def test_target_open(
    opener: Callable[[str | Path], Target], path: str, loader: type[Loader], mock_uac_dir: Path
) -> None:
    """Test that we correctly use the UAC loaders when opening a ``Target``."""
    path = mock_uac_dir if path == "mock_uac_dir" else absolute_path(path)

    with patch("dissect.target.target.Target.apply"):
        target = opener(path)

        assert isinstance(target._loader, loader)
        if isinstance(target._loader, TarLoader):
            assert isinstance(target._loader.subloader, UacTarSubloader)
        elif isinstance(target._loader, ZipLoader):
            assert isinstance(target._loader.subloader, UacZipSubLoader)
        assert target.path == path


@pytest.mark.parametrize(
    "data_path",
    [
        ("_data/loaders/uac/uac-2e44ea6da71d-linux-20250717143111.tar.gz"),
        # this one start with ./[root]
        ("_data/loaders/uac/uac-2e44ea6da71d-linux-20250717143112.tar.gz"),
    ],
)
def test_compressed_tar(data_path: str) -> None:
    """Test if we map a compressed UAC tar image correctly."""
    path = absolute_path(data_path)

    loader = loader_open(path)
    assert isinstance(loader, TarLoader)

    t = Target()
    loader.map(t)
    assert isinstance(loader.subloader, UacTarSubloader)
    assert len(t.filesystems) == 1

    t.apply()
    test_file = t.fs.path("/etc/passwd")
    assert test_file.exists()
    assert test_file.is_file()
    assert test_file.open().readline() == b"root:x:0:0:root:/root:/bin/bash\n"


def test_compressed_zip() -> None:
    """Test if we map a compressed UAC zip image correctly."""
    path = absolute_path("_data/loaders/uac/uac-2e44ea6da71d-linux-20250717143106.zip")

    loader = loader_open(path)
    assert isinstance(loader, ZipLoader)

    t = Target()
    loader.map(t)
    assert isinstance(loader.subloader, UacZipSubLoader)
    assert len(t.filesystems) == 1

    t.apply()
    test_file = t.fs.path("etc/passwd")
    assert test_file.exists()
    assert test_file.is_file()
    assert test_file.open().readline() == b"root:x:0:0:root:/root:/bin/bash\n"


def test_dir(mock_uac_dir: Path) -> None:
    """Test if we map an extracted UAC directory correctly."""

    loader = loader_open(mock_uac_dir)
    assert isinstance(loader, UacLoader)

    t = Target()
    loader.map(t)
    assert len(t.filesystems) == 1

    t.apply()
    test_file = t.fs.path("etc/passwd")
    assert test_file.exists()
    assert test_file.is_file()
    assert test_file.open().readline() == b"root:x:0:0:root:/root:/bin/bash\n"


@pytest.mark.parametrize(
    ("archive", "loader"),
    [
        ("_data/loaders/uac/uac-2e44ea6da71d-linux-20250717143111.tar.gz", TarLoader),
        ("_data/loaders/uac/uac-2e44ea6da71d-linux-20250717143106.zip", ZipLoader),
    ],
)
@pytest.mark.benchmark
def test_benchmark(benchmark: BenchmarkFixture, archive: str, loader: type[Loader]) -> None:
    file = absolute_path(archive)

    benchmark(lambda: loader(file).map(Target()))



#!/usr/bin/env python
from __future__ import annotations

import argparse
import dataclasses
import logging
import os
import re
import shutil
import sys
from difflib import diff_bytes, unified_diff
from fnmatch import fnmatch, translate
from io import BytesIO
from typing import TYPE_CHECKING, TextIO

from dissect.cstruct import hexdump
from flow.record import Record, RecordOutput, ignore_fields_for_comparison

from dissect.target.exceptions import FileNotFoundError
from dissect.target.helpers import fsutil
from dissect.target.helpers.logging import get_logger
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import alias, arg
from dissect.target.target import Target
from dissect.target.tools.query import record_output
from dissect.target.tools.shell import (
    ANSI_COLORS,
    ExtendedCmd,
    TargetCli,
    _target_name,
    arg_str_to_arg_list,
    build_pipe_stdout,
    fmt_ls_colors,
    python_shell,
    run_cli,
)
from dissect.target.tools.utils.cli import (
    catch_sigpipe,
    configure_generic_arguments,
    generate_argparse_for_method,
    process_generic_arguments,
)
from dissect.target.tools.utils.fs import print_extensive_file_stat_listing

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.filesystem import FilesystemEntry


log = get_logger(__name__)
logging.lastResort = None
logging.raiseExceptions = False

BLOCK_SIZE = 2048
FILE_LIMIT = BLOCK_SIZE * 16

FILE_DIFF_RECORD_FIELDS = [
    ("string", "src_target"),
    ("string", "dst_target"),
    ("string", "path"),
]
RECORD_DIFF_RECORD_FIELDS = [
    ("string", "src_target"),
    ("string", "dst_target"),
    ("record", "record"),
]

FileDeletedRecord = TargetRecordDescriptor("differential/file/deleted", FILE_DIFF_RECORD_FIELDS)
FileCreatedRecord = TargetRecordDescriptor("differential/file/created", FILE_DIFF_RECORD_FIELDS)
FileModifiedRecord = TargetRecordDescriptor(
    "differential/file/modified",
    [*FILE_DIFF_RECORD_FIELDS, ("bytes[]", "diff")],
)

RecordCreatedRecord = TargetRecordDescriptor("differential/record/created", RECORD_DIFF_RECORD_FIELDS)
RecordDeletedRecord = TargetRecordDescriptor("differential/record/deleted", RECORD_DIFF_RECORD_FIELDS)
RecordUnchangedRecord = TargetRecordDescriptor("differential/record/unchanged", RECORD_DIFF_RECORD_FIELDS)


@dataclasses.dataclass
class DifferentialEntry:
    """Signifies a change for a FilesystemEntry between two versions of a target."""

    path: str
    name: str
    src_target_entry: FilesystemEntry
    dst_target_entry: FilesystemEntry
    diff: list[bytes]


@dataclasses.dataclass
class DirectoryDifferential:
    """For a given directory, contains the unchanged, created, modified and deleted entries, as well as a list of
    subdirectories."""

    directory: str
    unchanged: list[FilesystemEntry] = dataclasses.field(default_factory=list)
    created: list[FilesystemEntry] = dataclasses.field(default_factory=list)
    modified: list[DifferentialEntry] = dataclasses.field(default_factory=list)
    deleted: list[FilesystemEntry] = dataclasses.field(default_factory=list)


def likely_unchanged(src: fsutil.stat_result, dst: fsutil.stat_result) -> bool:
    """Determine whether or not, based on the file stats, we can assume a file hasn't been changed."""
    return not (src.st_size != dst.st_size or src.st_mtime != dst.st_mtime or src.st_ctime != dst.st_ctime)


def get_plugin_output_records(plugin_name: str, plugin_arg_parts: list[str], target: Target) -> Iterator[Record]:
    """Command exection helper for target plugins.

    Highly similar to target-shell's ``_exec_target``, however this function
    only accepts plugins that outputs records, and returns an iterable of records rather than a function that outputs
    to stdout.
    """
    attr = target
    for part in plugin_name.split("."):
        attr = getattr(attr, part)

    if getattr(attr, "__output__", "default") != "record":
        raise ValueError("Comparing plugin output is only supported for plugins outputting records.")

    if callable(attr):
        argparser = generate_argparse_for_method(attr)
        try:
            args = argparser.parse_args(plugin_arg_parts)
        except SystemExit:
            return False

        return attr(**vars(args))

    return attr


class TargetComparison:
    """This class wraps functionality that for two given targets can identify similarities and differences between them.

    Currently supports differentiating between the target filesystems, and between plugin outputs.
    """

    def __init__(
        self,
        src_target: Target,
        dst_target: Target,
        deep: bool = False,
        file_limit: int = FILE_LIMIT,
    ):
        self.src_target = src_target
        self.dst_target = dst_target
        self.deep = deep
        self.file_limit = file_limit

    def scandir(self, path: str) -> DirectoryDifferential:
        """Scan a given directory for files that have been unchanged, modified, created or deleted from one target to
        the next. Add these results (as well as subdirectories) to a DirectoryDifferential object."""
        unchanged = []
        modified = []
        exists_as_directory_src = self.src_target.fs.exists(path) and self.src_target.fs.get(path).is_dir()
        exists_as_directory_dst = self.dst_target.fs.exists(path) and self.dst_target.fs.get(path).is_dir()

        # TODO: Adjust the following code to deal more efficiently with DirEntry instead of FilesystemEntry
        if not (exists_as_directory_src and exists_as_directory_dst):
            if exists_as_directory_src:
                # Path only exists on src target, hence all entries can be considered 'deleted'
                entries = [entry.get() for entry in self.src_target.fs.scandir(path)]
                return DirectoryDifferential(path, deleted=entries)
            if exists_as_directory_dst:
                # Path only exists on dst target, hence all entries can be considered 'created'
                entries = [entry.get() for entry in self.dst_target.fs.scandir(path)]
                return DirectoryDifferential(path, created=entries)
            raise ValueError(f"{path} is not a directory on either the source or destination target!")

        src_target_entries = [entry.get() for entry in self.src_target.fs.scandir(path)]
        src_target_children_paths = {entry.path for entry in src_target_entries}

        dst_target_entries = [entry.get() for entry in self.dst_target.fs.scandir(path)]
        dst_target_children_paths = {entry.path for entry in dst_target_entries}

        paths_only_on_src_target = src_target_children_paths - dst_target_children_paths
        paths_only_on_dst_target = dst_target_children_paths - src_target_children_paths

        deleted = [entry for entry in src_target_entries if entry.path in paths_only_on_src_target]
        created = [entry for entry in dst_target_entries if entry.path in paths_only_on_dst_target]

        paths_on_both = src_target_children_paths.intersection(dst_target_children_paths)
        entry_pairs = []

        for dst_entry in dst_target_entries:
            if dst_entry.path not in paths_on_both:
                continue
            src_entry = next((entry for entry in src_target_entries if entry.path == dst_entry.path), None)
            entry_pairs.append((src_entry, dst_entry))

        for entry_pair in entry_pairs:
            src_entry, dst_entry = entry_pair
            entry_path = src_entry.path

            # It's possible that there is an entry, but upon trying to retrieve its stats / content, we get a
            # FileNotFoundError. We account for this by wrapping both stat retrievals in a try except
            src_target_notfound = False
            dst_target_notfound = False
            src_target_isdir = None
            dst_target_isdir = None

            try:
                src_target_stat = src_entry.stat()
                src_target_isdir = src_entry.is_dir()
            except FileNotFoundError:
                src_target_notfound = True

            try:
                dst_target_stat = dst_entry.stat()
                dst_target_isdir = dst_entry.is_dir()
            except FileNotFoundError:
                dst_target_notfound = True

            if src_target_notfound or dst_target_notfound:
                if src_target_notfound and not dst_target_notfound:
                    created.append(dst_entry)
                elif dst_target_notfound and not src_target_notfound:
                    deleted.append(src_entry)
                else:
                    # Not found on both
                    unchanged.append(src_entry)
                # We can't continue as we cannot access the stats (or buffer)
                continue

            if src_target_isdir or dst_target_isdir:
                if src_target_isdir == dst_target_isdir:
                    unchanged.append(src_entry)
                else:
                    # Went from a file to a dir, or from a dir to a file. Either way, we consider the source entry
                    # 'deleted' and the dst entry 'Created'
                    deleted.append(src_entry)
                    created.append(dst_entry)
                continue

            if self.deep is False and likely_unchanged(src_target_stat, dst_target_stat):
                unchanged.append(src_entry)
                continue

            # If we get here, we have two files that we need to compare contents of
            src_fh = src_entry.open()
            dst_fh = dst_entry.open()

            # We reverse read the file object.
            chunk_size = 1024 * 10
            chunks_a = fsutil.reverse_read(src_fh, chunk_size, reverse_chunk=False)
            chunks_b = fsutil.reverse_read(dst_fh, chunk_size, reverse_chunk=False)

            chunk_count = 0

            while True:
                chunk_a = next(chunks_a, b"")
                chunk_b = next(chunks_b, b"")
                chunk_count += 1

                if chunk_a != chunk_b:
                    # We immediately break after discovering a difference in file contents
                    # This means that we won't return a full diff of the file, merely the first block where a difference
                    # is observed. The chunk is not reversed, so the difference is human-readable.
                    content_difference = list(diff_bytes(unified_diff, [chunk_a], [chunk_b]))
                    differential_entry = DifferentialEntry(
                        entry_path,
                        src_entry.name,
                        src_entry,
                        dst_entry,
                        content_difference,
                    )
                    modified.append(differential_entry)
                    break

                if self.file_limit and chunk_count * chunk_size > self.file_limit:
                    unchanged.append(src_entry)
                    break

                if not chunk_a:
                    # End of file
                    unchanged.append(src_entry)
                    break

        return DirectoryDifferential(path, unchanged, created, modified, deleted)

    def walkdir(
        self,
        path: str,
        exclude: list[str] | str | None = None,
        already_iterated: list[str] | None = None,
    ) -> Iterator[DirectoryDifferential]:
        """Recursively iterate directories and yield DirectoryDifferentials."""
        if already_iterated is None:
            already_iterated = []

        if path in already_iterated:
            return

        if exclude is not None and not isinstance(exclude, list):
            exclude = [exclude]

        already_iterated.append(path)

        # Do not scan the given path if it matches any excluded path.
        if exclude and next((pattern for pattern in exclude if fnmatch(path, pattern)), None):
            return

        diff = self.scandir(path)

        # Check if diff contains excluded paths
        if exclude:
            for t in ["created", "unchanged", "modified", "deleted"]:
                for i, d in enumerate(getattr(diff, t)):
                    if next((pattern for pattern in exclude if fnmatch(d.path, pattern)), None):
                        del getattr(diff, t)[i]
        yield diff

        subentries = diff.created + diff.unchanged + diff.deleted
        subdirectories = [entry for entry in subentries if entry.is_dir()]
        # Check if the scandir lead to the discovery of new directories that we have to scan for differentials
        # Directories are always in 'unchanged'
        for subdirectory in subdirectories:
            if subdirectory in already_iterated:
                continue

            # Right-pad with a '/'
            subdirectory_path = subdirectory.path if subdirectory.path.endswith("/") else subdirectory.path + "/"
            if exclude:
                match = next((pattern for pattern in exclude if fnmatch(subdirectory_path, pattern)), None)
                if match:
                    continue
            yield from self.walkdir(subdirectory.path, exclude, already_iterated)

    def differentiate_plugin_outputs(
        self, plugin_name: str, plugin_arg_parts: list[str], only_changed: bool = False
    ) -> Iterator[Record]:
        """Run a plugin on the source and destination targets and yield RecordUnchanged, RecordCreated and RecordDeleted
        records.

        There is no equivalent for the FileModifiedRecord. For files and directories, we can use the path to
        reliably track changes from one target to the next. There is no equivalent for plugin outputs, so we just assume
        that all records are either deleted (only on src), created (only on dst) or unchanged (on both).
        """
        with ignore_fields_for_comparison(["_generated", "_source", "hostname", "domain"]):
            src_records = set(get_plugin_output_records(plugin_name, plugin_arg_parts, self.src_target))
            src_records_seen = set()

            for dst_record in get_plugin_output_records(plugin_name, plugin_arg_parts, self.dst_target):
                if dst_record in src_records:
                    src_records_seen.add(dst_record)
                    if not only_changed:
                        yield RecordUnchangedRecord(
                            src_target=self.src_target.path, dst_target=self.dst_target.path, record=dst_record
                        )
                else:
                    yield RecordCreatedRecord(
                        src_target=self.src_target.path, dst_target=self.dst_target.path, record=dst_record
                    )
            for record in src_records - src_records_seen:
                yield RecordDeletedRecord(
                    src_target=self.src_target.path, dst_target=self.dst_target.path, record=record
                )


class DifferentialCli(ExtendedCmd):
    """CLI for browsing the differential between two or more targets."""

    doc_header_prefix = "target-diff\n==========\n"
    doc_header_suffix = "\n\nDocumented commands (type help <topic>):"
    doc_header_multiple_targets = "Use 'list', 'prev' and 'next' to list and select targets to differentiate between."

    def __init__(self, *targets: tuple[Target], deep: bool = False, limit: int = FILE_LIMIT):
        self.targets = targets
        self.deep = deep
        self.limit = limit

        self.src_index = 0
        self.dst_index = 0
        self.comparison: TargetComparison = None

        self.cwd = "/"
        self.alt_separator = "/"

        doc_header_middle = self.doc_header_multiple_targets if len(targets) > 2 else ""
        self.doc_header = self.doc_header_prefix + doc_header_middle + self.doc_header_suffix

        self._select_source_and_dest(0, 1)

        start_in_cyber = any(target.props.get("cyber") for target in self.targets)
        super().__init__(start_in_cyber)

        if len(self.targets) > 2:
            # Some help may be nice if you are diffing more than 2 targets at once
            self.do_help(arg=None)

    @property
    def src_target(self) -> Target:
        return self.targets[self.src_index]

    @property
    def dst_target(self) -> Target:
        return self.targets[self.dst_index]

    @property
    def prompt(self) -> str:
        """Determine the prompt of the cli."""

        src_name = _target_name(self.comparison.src_target)
        dst_name = _target_name(self.comparison.dst_target)

        base = f"{src_name}/{dst_name}" if src_name != dst_name else src_name
        suffix = "{base}:{cwd}$" if os.getenv("NO_COLOR") else "{BOLD_GREEN}{base}{RESET}:{BOLD_BLUE}{cwd}{RESET}$"
        suffix = suffix.format(base=base, cwd=self.cwd, **ANSI_COLORS)  # type: ignore

        if len(self.targets) <= 2:
            return f"(diff) {suffix} "

        chain_prefix = "[ "
        for i in range(len(self.targets)):
            char = "O " if i == self.src_index or i == self.dst_index else ". "
            chain_prefix += char
        chain_prefix += "] "

        return f"(diff) {chain_prefix}{suffix} "

    def _select_source_and_dest(self, src_index: int, dst_index: int) -> None:
        """Set local variables according to newly selected source and destination index, and re-instatiate
        TargetComparison."""
        self.src_index = src_index
        self.dst_index = dst_index
        if not self.src_target.fs.exists(self.cwd) and not self.dst_target.fs.exists(self.cwd):
            logging.warning("The current directory exists on neither of the selected targets")
        if self.src_target.fs.alt_separator != self.dst_target.fs.alt_separator:
            raise NotImplementedError("No support for handling targets with different path separators")

        self.alt_separator = self.src_target.fs.alt_separator
        self.comparison = TargetComparison(self.src_target, self.dst_target, self.deep, self.limit)

    def _annotate_differential(
        self,
        diff: DirectoryDifferential,
        unchanged: bool = True,
        created: bool = True,
        modified: bool = True,
        deleted: bool = True,
        absolute: bool = False,
    ) -> list[tuple[fsutil.TargetPath | DifferentialEntry], str]:
        """Given a DirectoryDifferential instance, construct a list of tuples where the first element is a Filesystem /
        DifferentialEntry and the second a color-formatted string."""
        r = []

        attr = "path" if absolute else "name"
        if unchanged:
            for entry in diff.unchanged:
                color = "di" if entry.is_dir() else "fi"
                r.append((entry, fmt_ls_colors(color, getattr(entry, attr))))

        if created:
            for entry in diff.created:
                color = "tw" if entry.is_dir() else "ex"
                r.append((entry, fmt_ls_colors(color, f"{getattr(entry, attr)} (created)")))

        if modified:
            for entry in diff.modified:
                # Modified entries are always files
                r.append((entry, fmt_ls_colors("ln", f"{getattr(entry, attr)} (modified)")))  # noqa: PERF401
        if deleted:
            for entry in diff.deleted:
                color = "su" if entry.is_dir() else "or"
                r.append((entry, fmt_ls_colors(color, f"{getattr(entry, attr)} (deleted)")))

        r.sort(key=lambda e: e[0].name)
        return r

    def _targets_with_directory(self, path: str, warn_when_incomplete: bool = False) -> int:
        """Return whether a given path is an existing directory for neither, one of, or both of the targets being
        compared. Optionally log a warning if the directory only exists on one of the two targets."""
        src_has_dir = False
        dst_has_dir = False
        try:
            entry = self.comparison.src_target.fs.get(path)
            src_has_dir = entry.is_dir()
        except FileNotFoundError:
            pass
        try:
            entry = self.comparison.dst_target.fs.get(path)
            dst_has_dir = entry.is_dir()
        except FileNotFoundError:
            pass

        if (src_has_dir is False or dst_has_dir is False) and warn_when_incomplete:
            if src_has_dir != dst_has_dir:
                target_with_dir = self.comparison.src_target if src_has_dir else self.comparison.dst_target
                log.warning("%r is only a valid path on '%s'", path, target_with_dir)
            else:
                log.warning("%r is not a valid path on either target", path)
        return int(src_has_dir) + int(dst_has_dir)

    def _write_entry_contents_to_stdout(self, entry: FilesystemEntry, stdout: TextIO) -> bool:
        """Copy the contents of a Filesystementry to stdout."""
        stdout = stdout.buffer
        fh = entry.open()
        shutil.copyfileobj(fh, stdout)
        stdout.flush()
        print()
        return False

    def completedefault(self, text: str, line: str, begidx: int, endidx: int) -> list[str]:
        """Autocomplete based on files / directories found in the current path."""
        path = line[:begidx].rsplit(" ")[-1]
        textlower = text.lower()

        path = fsutil.abspath(path, cwd=str(self.cwd), alt_separator=self.alt_separator)

        diff = self.comparison.scandir(path)
        items = [
            (item.entry.is_dir(), item.name) for group in [diff.created, diff.unchanged, diff.deleted] for item in group
        ]
        items += [
            (item.src_target_entry.is_dir() and item.dst_target_entry.is_dir(), item.name) for item in diff.modified
        ]
        suggestions = []
        for is_dir, fname in items:
            if not fname.lower().startswith(textlower):
                continue

            # Add a trailing slash to directories, to allow for easier traversal of the filesystem
            suggestion = f"{fname}/" if is_dir else fname
            suggestions.append(suggestion)
        return suggestions

    def do_list(self, line: str) -> bool:
        """print a list of targets to differentiate between

        Useful when differentiating between three or more targets. Looks quite bad on small terminal screens.
        """
        columns = ["#", "Name", "Path", "From", "To"]

        rows = []

        for i, target in enumerate(self.targets):
            rows.append(
                [
                    f"{i:2d}",
                    target.name,
                    str(target.path),
                    "**" if i == self.src_index else "",
                    "**" if i == self.dst_index else "",
                ]
            )

        longest_name = max(len(row[1]) + 4 for row in rows)
        longest_path = max(len(row[2]) + 4 for row in rows)
        name_len = max(10, longest_name)
        path_len = max(15, longest_path)

        fmt = "{:^5} | {:<" + str(name_len) + "} | {:<" + str(path_len) + "} | {:^6} | {:^6} |"
        print(fmt.format(*columns))
        print()
        for row in rows:
            print(fmt.format(*row))
            print()
        return False

    @alias("prev")
    @arg("-a", "--absolute", action="store_true", help="Only move the destination target one position back.")
    def cmd_previous(self, args: argparse.Namespace, line: str) -> bool:
        """when three or more targets are available, move the 'comparison window' one position back"""
        src_index = self.src_index - 1 if not args.absolute else 0
        if src_index < 0:
            src_index = len(self.targets) - 1
        dst_index = self.dst_index - 1
        if dst_index < 0:
            dst_index = len(self.targets) - 1
        if dst_index <= src_index:
            src_index, dst_index = dst_index, src_index
        self._select_source_and_dest(src_index, dst_index)
        return False

    @arg("-a", "--absolute", action="store_true", help="Only move the destination target one position forward.")
    def cmd_next(self, args: argparse.Namespace, line: str) -> bool:
        """when three or more targets are available, move the 'comparison window' one position forward"""
        dst_index = (self.dst_index + 1) % len(self.targets)
        src_index = self.src_index + 1 % len(self.targets) if not args.absolute else 0

        if dst_index <= src_index:
            src_index, dst_index = dst_index, src_index
        self._select_source_and_dest(src_index, dst_index)
        return False

    def do_cd(self, path: str) -> bool:
        """change directory to the given path"""
        path = fsutil.abspath(path, cwd=str(self.cwd), alt_separator=self.alt_separator)
        if self._targets_with_directory(path, warn_when_incomplete=True) != 0:
            self.cwd = path
        return False

    @arg("path", nargs="?")
    @arg("-l", action="store_true")
    @arg("-a", "--all", action="store_true")  # ignored but included for proper argument parsing
    @arg("-h", "--human-readable", action="store_true")
    def cmd_ls(self, args: argparse.Namespace, stdout: TextIO) -> bool:
        """list contents of a directory for two targets"""
        path = args.path if args.path is not None else self.cwd
        diff = self.comparison.scandir(path)
        results = self._annotate_differential(diff)
        if not args.l:
            print("\n".join([name for _, name in results]), file=stdout)
        else:
            for entry, name in results:
                if not isinstance(entry, DifferentialEntry):
                    print_extensive_file_stat_listing(stdout, name, entry, human_readable=args.human_readable)
                else:
                    # We have to choose for which version of this file we are going to print detailed info. The
                    # destination target seems to make the most sense: it is likely newer
                    print_extensive_file_stat_listing(
                        stdout, name, entry.dst_target_entry, human_readable=args.human_readable
                    )
        return False

    @arg("path", nargs="?")
    def cmd_cat(self, args: argparse.Namespace, stdout: TextIO) -> bool:
        """output the contents of a file"""
        base_dir, _, name = args.path.rpartition("/")
        if not base_dir:
            base_dir = self.cwd

        directory_differential = self.comparison.scandir(base_dir)
        entry = None
        for entry in directory_differential.unchanged:
            if entry.name == name:
                return self._write_entry_contents_to_stdout(entry, stdout)
        for entry in directory_differential.created:
            if entry.name == name:
                log.warning("%r is only present on '%s'", entry.name, self.comparison.dst_target.path)
                return self._write_entry_contents_to_stdout(entry, stdout)
        for entry in directory_differential.deleted:
            if entry.name == name:
                log.warning("%r is only present on '%s'", entry.name, self.comparison.src_target.path)
                return self._write_entry_contents_to_stdout(entry, stdout)
        for entry in directory_differential.modified:
            if entry.name == name:
                log.warning(
                    "Concatinating latest version of '%s'. Use 'diff' to differentiate between target versions.",
                    entry.name,
                )
                return self._write_entry_contents_to_stdout(entry.dst_target_entry, stdout)
        print(f"File {name} not found.")
        return False

    @arg("path", nargs="?")
    @arg("--hex", action="store_true")
    def cmd_diff(self, args: argparse.Namespace, stdout: TextIO) -> bool:
        """output the difference in file contents between two targets"""
        stdout = stdout.buffer
        base_dir, _, name = args.path.rpartition("/")
        if not base_dir:
            base_dir = self.cwd
        directory_differential = self.comparison.scandir(base_dir)
        for entry in directory_differential.modified:
            if entry.name == name:
                if args.hex:
                    primary_fh_lines = [
                        line.encode()
                        for line in hexdump(entry.src_target_entry.open().read(), output="string").split("\n")
                    ]
                    secondary_fh_lines = [
                        line.encode()
                        for line in hexdump(entry.dst_target_entry.open().read(), output="string").split("\n")
                    ]
                else:
                    primary_fh_lines = entry.src_target_entry.open().readlines()
                    secondary_fh_lines = entry.dst_target_entry.open().readlines()

                for chunk in diff_bytes(unified_diff, primary_fh_lines, secondary_fh_lines):
                    if chunk.startswith(b"@@"):
                        chunk = fmt_ls_colors("ln", chunk.decode()).encode()
                    elif chunk.startswith(b"+"):
                        chunk = fmt_ls_colors("ex", chunk.decode()).encode()
                    elif chunk.startswith(b"-"):
                        chunk = fmt_ls_colors("or", chunk.decode()).encode()

                    shutil.copyfileobj(BytesIO(chunk), stdout)

                    if args.hex:
                        stdout.write(b"\n")

                    stdout.flush()

                print()
                return False

        # Check if this file is even present on one of the targets
        files = directory_differential.unchanged + directory_differential.created + directory_differential.deleted
        match = next((entry for entry in files if entry.name == name), None)
        if match is None:
            print(f"File {name} not found.")
        else:
            print(f"No two versions available for {name} to differentiate between.")
        return False

    @arg("path", nargs="?")
    @alias("xxd")
    def cmd_hexdump(self, args: argparse.Namespace, stdout: TextIO) -> bool:
        """output difference of the given file between targets in hexdump"""
        args.hex = True
        return self.cmd_diff(args, stdout)

    @arg("index")
    @arg("type", choices=["src", "dst"])
    def cmd_set(self, args: argparse.Namespace, stdout: TextIO) -> bool:
        """change either the source or destination target for differentiation

        Index can be given relative (when prefixed with '+' or '-', e.g. "set dst +1") or absolute (e.g. set src 0).
        """
        index = args.index.strip()
        pos = self.src_index if args.type == "src" else self.dst_index

        if index.startswith(("+", "-")):
            multiplier = 1 if index[0] == "+" else -1
            index = index[1:].strip()
            if not index.isdigit():
                return False
            pos += int(index) * multiplier
        elif index.isdigit():
            pos = int(index)
        else:
            raise ValueError(f"Could not set {args.type} to {index}.")
        if args.type == "src":
            self._select_source_and_dest(pos, self.dst_index)
        else:
            self._select_source_and_dest(self.src_index, pos)
        return False

    @arg("target", choices=["src", "dst"])
    def cmd_enter(self, args: argparse.Namespace, stdout: TextIO) -> bool:
        """open a subshell for the source or destination target"""
        target = self.src_target if args.target == "src" else self.dst_target
        cli = TargetCli(target)
        if target.fs.exists(self.cwd):
            cli.chdir(self.cwd)

        # Cyber doesn't work well with subshells
        cli.cyber = False
        run_cli(cli)
        return False

    @arg("path", nargs="?")
    @arg("-name", default="*")
    @arg("-iname")
    @arg("-c", "--created", action="store_true")
    @arg("-m", "--modified", action="store_true")
    @arg("-d", "--deleted", action="store_true")
    @arg("-u", "--unchanged", action="store_true")
    def cmd_find(self, args: argparse.Namespace, stdout: TextIO) -> bool:
        """search for files in a directory hierarchy"""
        path = fsutil.abspath(args.path, cwd=str(self.cwd), alt_separator=self.comparison.src_target.fs.alt_separator)
        if not path:
            return False

        if self._targets_with_directory(path, warn_when_incomplete=True) == 0:
            return False

        pattern = re.compile(translate(args.iname), re.IGNORECASE) if args.iname else re.compile(translate(args.name))

        include_all_changes = not (args.created or args.modified or args.deleted or args.unchanged)

        include_unchanged = args.unchanged
        include_modified = include_all_changes or args.modified
        include_created = include_all_changes or args.created
        include_deleted = include_all_changes or args.deleted

        for differential in self.comparison.walkdir(path):
            for entry, line in self._annotate_differential(
                differential, include_unchanged, include_created, include_modified, include_deleted, absolute=True
            ):
                if not pattern.match(entry.name):
                    continue

                print(line, file=stdout)

        return False

    def do_plugin(self, line: str) -> bool:
        """yield RecordCreated, RecordUnchanged and RecordDeleted Records by comparing plugin outputs for two targets"""
        argparts = arg_str_to_arg_list(line)
        pipeparts = []
        if "|" in argparts:
            pipeidx = argparts.index("|")
            argparts, pipeparts = argparts[:pipeidx], argparts[pipeidx + 1 :]

        if len(argparts) < 1:
            raise ValueError("Provide a plugin name, and optionally parameters to pass to the plugin.")

        plugin = argparts.pop(0)

        iterator = self.comparison.differentiate_plugin_outputs(plugin, argparts)
        if pipeparts:
            try:
                with build_pipe_stdout(pipeparts) as pipe_stdin:
                    rs = RecordOutput(pipe_stdin.buffer)
                    for record in iterator:
                        rs.write(record)
            except OSError as e:
                # in case of a failure in a subprocess
                print(e)
        else:
            for record in iterator:
                print(record, file=sys.stdout)

        return False

    def do_python(self, line: str) -> bool:
        """drop into a Python shell"""
        python_shell(list(self.targets))
        return False


def make_target_pairs(targets: tuple[Target], absolute: bool = False) -> list[tuple[Target, Target]]:
    """Make 'pairs' of targets that we are going to compare against one another.

    A list of targets can be treated in two
    ways: compare every target with the one that came before it, or compare all targets against a 'base' target
    (which has to be supplied as initial target in the list).
    """
    target_pairs = []

    previous_target = targets[0]
    for target in targets[1:]:
        target_pairs.append((previous_target, target))
        if not absolute:
            # The next target should be compared against the one we just opened
            previous_target = target
    return target_pairs


def differentiate_target_filesystems(
    *targets: tuple[Target],
    deep: bool = False,
    limit: int = FILE_LIMIT,
    absolute: bool = False,
    include: list[str] | None = None,
    exclude: list[str] | None = None,
) -> Iterator[Record]:
    """Given a list of targets, compare targets against one another and yield File[Created|Modified|Deleted]Records
    indicating the differences between them.
    """

    for target_pair in make_target_pairs(targets, absolute):
        # Unpack the tuple and initialize the comparison class
        src_target, dst_target = target_pair
        comparison = TargetComparison(src_target, dst_target, deep, limit)

        paths = ["/"] if include is None else include

        for path in paths:
            for directory_diff in comparison.walkdir(path, exclude=exclude):
                for creation_entry in directory_diff.created:
                    yield FileCreatedRecord(
                        path=creation_entry.path,
                        src_target=src_target.path,
                        dst_target=dst_target.path,
                    )

                for deletion_entry in directory_diff.deleted:
                    yield FileDeletedRecord(
                        path=deletion_entry.path,
                        src_target=src_target.path,
                        dst_target=dst_target.path,
                    )

                for entry_difference in directory_diff.modified:
                    yield FileModifiedRecord(
                        path=entry_difference.path,
                        diff=entry_difference.diff,
                        src_target=src_target.path,
                        dst_target=dst_target.path,
                    )


def differentiate_target_plugin_outputs(
    *targets: tuple[Target], absolute: bool = False, only_changed: bool = False, plugin: str, plugin_args: str = ""
) -> Iterator[Record]:
    """Given a list of targets, yielding records indicating which records from this plugin are new, unmodified or
    deleted."""
    for target_pair in make_target_pairs(targets, absolute):
        src_target, dst_target = target_pair
        comparison = TargetComparison(src_target, dst_target)
        yield from comparison.differentiate_plugin_outputs(plugin, plugin_args, only_changed)


@catch_sigpipe
def main() -> int:
    help_formatter = argparse.ArgumentDefaultsHelpFormatter
    parser = argparse.ArgumentParser(
        description="target-diff",
        fromfile_prefix_chars="@",
        formatter_class=help_formatter,
    )

    parser.add_argument(
        "-d",
        "--deep",
        action="store_true",
        help="compare file contents even if metadata suggests they have been left unchanged",
    )
    parser.add_argument(
        "-l",
        "--limit",
        default=FILE_LIMIT,
        type=int,
        help="how many bytes to compare before assuming a file is left unchanged (0 for no limit)",
    )
    subparsers = parser.add_subparsers(help="mode for differentiating targets", dest="mode", required=True)

    shell_mode = subparsers.add_parser("shell", help="open an interactive shell to compare two or more targets.")
    shell_mode.add_argument("targets", metavar="TARGETS", nargs="+", help="targets to differentiate between")

    fs_mode = subparsers.add_parser("fs", help="yield records about differences between target filesystems.")
    fs_mode.add_argument("targets", metavar="TARGETS", nargs="+", help="targets to differentiate between")
    fs_mode.add_argument("-s", "--strings", action="store_true", help="print records as strings")
    fs_mode.add_argument("-e", "--exclude", action="append", help="path(s) on targets not to check for differences")
    fs_mode.add_argument(
        "-i",
        "--include",
        action="append",
        help="path(s) on targets to check for differences (all will be checked if left omitted)",
    )
    fs_mode.add_argument(
        "-a",
        "--absolute",
        action="store_true",
        help=(
            "treat every target as an absolute. The first given target is treated as the 'base' target to compare "
            "subsequent targets against. If omitted, every target is treated as a 'delta' and compared against the "
            "target that came before it."
        ),
    )

    query_mode = subparsers.add_parser("query", help="differentiate plugin outputs between two or more targets.")
    query_mode.add_argument("targets", metavar="TARGETS", nargs="+", help="targets to differentiate between")
    query_mode.add_argument("-s", "--strings", action="store_true", help="print records as strings")
    query_mode.add_argument(
        "-p",
        "--parameters",
        default="",
        help="parameters for the plugin",
    )
    query_mode.add_argument(
        "-f",
        "--plugin",
        required=True,
        help="function to execute",
    )
    query_mode.add_argument(
        "-a",
        "--absolute",
        action="store_true",
        help=(
            "treat every target as an absolute. The first given target is treated as the 'base' target to compare "
            "subsequent targets against. If omitted, every target is treated as a 'delta' and compared against the "
            "target that came before it."
        ),
    )
    query_mode.add_argument(
        "--only-changed",
        action="store_true",
        help="do not output unchanged records",
    )

    configure_generic_arguments(parser)

    args, _ = parser.parse_known_args()
    process_generic_arguments(parser, args)

    if len(args.targets) < 2:
        parser.error("at least two targets are required for target-diff")

    target_list = [Target.open(path) for path in args.targets]
    if args.mode == "shell":
        cli = DifferentialCli(*target_list, deep=args.deep, limit=args.limit)
        run_cli(cli)
    else:
        writer = record_output(args.strings)
        if args.mode == "fs":
            iterator = differentiate_target_filesystems(
                *target_list,
                deep=args.deep,
                limit=args.limit,
                absolute=args.absolute,
                include=args.include,
                exclude=args.exclude,
            )
        elif args.mode == "query":
            if args.deep:
                log.error("Argument --deep is not available in target-diff query mode")
                return 1

            if args.limit != FILE_LIMIT:
                log.error("Argument --limit is not available in target-diff query mode")
                return 1

            iterator = differentiate_target_plugin_outputs(
                *target_list,
                absolute=args.absolute,
                only_changed=args.only_changed,
                plugin=args.plugin,
                plugin_args=arg_str_to_arg_list(args.parameters),
            )

        try:
            for record in iterator:
                writer.write(record)

        except Exception as e:
            log.error(e)  # noqa: TRY400
            log.debug("", exc_info=e)
            return 1

    return 0


if __name__ == "__main__":
    main()



from __future__ import annotations

# Manual overrides for types freedesktop does not (yet) support.
TYPES = [
    {
        "type": "application/x-compress",
        "name": "UNIX-compressed file (LZH)",
        "pattern": ["*.Z", "*.z", "*.tar.z"],
        "magic": [{"offset": 0, "value": b"\x1f\xa0"}],
    },
    {
        "type": "application/x-plist",
        "name": "Apple binary property list file",
        "pattern": ["*.plist"],
        "magic": [{"offset": 0, "value": b"bplist"}],
    },
    {
        "type": "image/bigtiff",
        "name": "BigTIFF image",
        "pattern": ["*.tif", "*.tiff"],
        "magic": [
            {"offset": 0, "value": b"MM\x00+"},
            {"offset": 0, "value": b"II+\x00"},
        ],
    },
    {
        "type": "image/bpg",
        "name": "Better Portable Graphics image",
        "pattern": ["*.bpg"],
        "magic": [
            {"offset": 0, "value": b"\x42\x50\x47\xfb"},
        ],
    },
    {
        "type": "image/jp2",
        "name": "JPEG-2000 JP2 image",
        "pattern": ["*.jp2", "*.jpg2"],
        "magic": [{"offset": 0, "value": b"\x00\x00\x00\x0cjP  \r\n\x87\n"}],
    },
    {
        "type": "application/x-cpio",
        "name": "CPIO archive",
        "pattern": ["*.cpio"],
        "magic": [
            {"offset": 0, "value": b"070707"},
            {"offset": 0, "value": b"070701"},
            {"offset": 0, "value": b"070702"},
        ],
    },
    {
        "type": "application/zip",
        "name": "Zip archive",
        "pattern": ["*.zip", "*.zipx"],
        "magic": [
            {"offset": 0, "value": b"PK\x03\x04"},
            {"offset": 0, "value": b"PK\x05\x06"},
            {"offset": 0, "value": b"PK\x07\x08"},
        ],
    },
    {
        "type": "application/x-executable",
        "name": "Executable",
        "pattern": [],
        "magic": [
            {"offset": 0, "value": b"\x7fELF"},
            {"offset": 0, "value": b"R\x1c"},
            {"offset": 0, "value": b"\x04 "},
            {"offset": 0, "value": b"\x04!"},
            {"offset": 0, "value": b"\x06\x03"},
        ],
    },
    {
        "type": "application/x-mach-o",
        "name": "Mach-O executable",
        "pattern": [],
        "magic": [
            {"offset": 0, "value": b"\xfe\xed\xfa\xce"},
            {"offset": 0, "value": b"\xfe\xed\xfa\xcf"},
            {"offset": 0, "value": b"\xce\xfa\xed\xfe"},
            {"offset": 0, "value": b"\xcf\xfa\xed\xfe"},
        ],
    },
    {
        "type": "application/vnd.tcpdump.pcap",
        "name": "Network packet capture",
        "pattern": [
            "*.pcap",
        ],
        "magic": [
            {"offset": 0, "value": b"\xa1\xb2\xc3\xd4"},
            {"offset": 0, "value": b"\xd4\xc3\xb2\xa1"},
            {"offset": 0, "value": b"\xa1\xb2\x3c\x4d"},
            {"offset": 0, "value": b"\x4d\x3c\xb2\x4d"},
            {"offset": 0, "value": b"\xd4\xc3\xb2\xa1"},
            {"offset": 0, "value": b"\x4d\x3c\xb2\xa1"},
        ],
    },
    {
        "type": "application/vnd.ms-htmlhelp",
        "name": "CHM document",
        "pattern": ["*.chm"],
        "magic": [
            {"offset": 0, "value": b"\x49\x54\x53\x46\x03\x00\x00\x00\x60\x00\x00\x00"},
        ],
    },
    {
        "type": "application/vnd.efi.iso",
        "name": "Raw CD image",
        "pattern": ["*.iso", "*.iso9660"],
        "magic": [
            {"offset": 0x8001, "value": b"CD001"},
        ],
    },
    {
        "type": "application/vnd.efi.iso+compressed",
        "name": "Compressed ISO image",
        "pattern": ["*.isz"],
        "magic": [
            {"offset": 0, "value": b"IsZ!"},
        ],
    },
    {
        "type": "application/zstd",
        "name": "Zstandard archive",
        "pattern": ["*.zst"],
        "magic": [
            {"offset": 0, "value": b"\xfd/\xb5("},
            {"offset": 0, "value": b"\x28\xb5\x2f\xfd"},
        ],
    },
    {
        "type": "application/x-lz4",
        "name": "LZ4 archive",
        "pattern": ["*.lz4"],
        "magic": [
            {"offset": 0, "value": b'\x18M"\x04'},
            {"offset": 0, "value": b"\x18L!\x02"},
            {"offset": 0, "value": b"\x04\x22\x4d\x18"},
        ],
    },
    {
        "type": "application/zlib",
        "name": "Zlib archive",
        "pattern": ["*.zz"],
        "magic": [
            {"offset": 0, "value": b"\x78\x01"},
            {"offset": 0, "value": b"\x78\x5e"},
            {"offset": 0, "value": b"\x78\x9c"},
            {"offset": 0, "value": b"\x78\xda"},
            {"offset": 0, "value": b"\x78\x20"},
            {"offset": 0, "value": b"\x78\x7d"},
            {"offset": 0, "value": b"\x78\xbb"},
            {"offset": 0, "value": b"\x78\xf9"},
        ],
    },
    {
        "type": "application/avro",
        "name": "Apache Avro Object Container File",
        "pattern": ["*.avro"],
        "magic": [
            {"offset": 0, "value": b"Obj\x01"},
        ],
    },
    {
        "type": "application/vnd.ms-tape",
        "name": "Microsoft Tape Format (MTF)",
        "pattern": [],
        "magic": [
            {"offset": 0, "value": b"TAPE"},
        ],
    },
    {
        "type": "application/x-duckdb",
        "name": "DuckDB database file",
        "pattern": [],
        "magic": [
            {"offset": 0, "value": b"DUCK"},
        ],
    },
    {
        "type": "application/vnd.ms-outlook-pst",
        "name": "Microsoft Outlook Personal Storage Table file",
        "pattern": [],
        "magic": [
            {"offset": 0, "value": b"!BDN"},
        ],
    },
    {
        "type": "application/x-lua",
        "name": "Lua bytecode",
        "pattern": ["*.luac"],
        "magic": [
            {"offset": 0, "value": b"\x1bLua"},
        ],
    },
    {
        "type": "application/x-aff",
        "name": "Advanced Forensics Format",
        "pattern": ["*.aff"],
        "magic": [
            {"offset": 0, "value": b"AFF"},
        ],
    },
    {
        "type": "application/x-encase",
        "name": "EnCase EWF version 2 format",
        "pattern": ["*.e01", "*.E01"],
        "magic": [
            {"offset": 0, "value": b"EVF2"},
        ],
    },
    {
        "type": "application/x-encase",
        "name": "EnCase EWF version 1 format",
        "pattern": ["*.e01", "*.E01"],
        "magic": [
            {"offset": 0, "value": b"EVF"},
        ],
    },
    {
        "type": "application/x-oracle-virtualbox-vdi",
        "name": "Oracle VirtualBox Virtual Hard Disk file format",
        "pattern": ["*.vdi"],
        "magic": [
            {"offset": 0, "value": b"<<< Oracle VM VirtualBox Disk Image >>>"},
        ],
    },
    {
        "type": "application/x-win-evt",
        "name": "Windows Event Viewer file format",
        "pattern": ["*.evt"],
        "magic": [
            {"offset": 0, "value": b"LfLe"},
        ],
    },
    {
        "type": "application/x-win-evtx",
        "name": "Windows Event Viewer XML file format",
        "pattern": ["*.evtx"],
        "magic": [
            {"offset": 0, "value": b"ElfFile"},
        ],
    },
    {
        "type": "application/x-win-regf",
        "name": "Windows Registry file",
        "pattern": [],
        "magic": [
            {"offset": 0, "value": b"regf"},
        ],
    },
    {
        "type": "application/x-win-zonetransfer",
        "name": "Microsoft Zone Identifier for URL Security Zones",
        "pattern": ["*.Identifier"],
        "magic": [
            {"offset": 0, "value": b"[ZoneTransfer]"},
        ],
    },
    {
        "type": "application/x-systemd-journald-log",
        "name": "Systemd Journald log file",
        "pattern": ["*.journal"],
        "magic": [
            {"offset": 0, "value": b"LPKSHHRH"},
        ],
    },
    {
        "type": "application/x-ext-jbd-fs-journal",
        "name": "Linux EXT3/EXT4 jbd2 filesystem journal",
        "pattern": [],
        "magic": [
            {"offset": 0, "value": b"\xc0\x3b\x39\x98"},
        ],
    },
    {
        "type": "text/pkix-csr",
        "name": "X.509 certificate signing request",
        "pattern": ["*.csr"],
        "magic": [
            {"offset": 0, "value": b"-----BEGIN CERTIFICATE REQUEST-----"},
        ],
    },
    {
        "type": "text/x-ssh-private-key",
        "name": "OpenSSH private key",
        "pattern": [],
        "magic": [
            {"offset": 0, "value": b"-----BEGIN OPENSSH PRIVATE KEY-----"},
            {"offset": 0, "value": b"-----BEGIN PRIVATE KEY-----"},
            {"offset": 0, "value": b"-----BEGIN DSA PRIVATE KEY-----"},
            {"offset": 0, "value": b"-----BEGIN RSA PRIVATE KEY-----"},
        ],
    },
    {
        "type": "text/x-ssh-public-key",
        "name": "OpenSSH public key",
        "pattern": ["*.pub"],
        "magic": [
            {"offset": 0, "value": b"ssh-ed25519 "},
            {"offset": 0, "value": b"sk-ssh-ed25519@openssh.com "},
            {"offset": 0, "value": b"ecdsa-sha2-nistp521 "},
            {"offset": 0, "value": b"ecdsa-sha2-nistp384 "},
            {"offset": 0, "value": b"ecdsa-sha2-nistp256 "},
            {"offset": 0, "value": b"sk-ecdsa-sha2-nistp256@openssh.com "},
            {"offset": 0, "value": b"ssh-rsa "},
            {"offset": 0, "value": b"ssh-dss "},
            {"offset": 0, "value": b"-----BEGIN SSH2 PUBLIC KEY-----"},
        ],
    },
    {
        "type": "text/x-putty-private-key",
        "name": "PuTTY private key",
        "pattern": [],
        "magic": [
            {"offset": 0, "value": b"PuTTY-User-Key-File-2:"},
            {"offset": 0, "value": b"PuTTY-User-Key-File-3:"},
        ],
    },
]



from __future__ import annotations

import re

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.plugin import OperatingSystem
from dissect.target.plugins.apps.webserver.apache import (
    RE_ACCESS_COMMON_PATTERN,
    RE_REFERER_USER_AGENT_PATTERN,
    RE_REMOTE_PATTERN,
    RE_RESPONSE_TIME_PATTERN,
    ApachePlugin,
    LogFormat,
)

LOG_FORMAT_CITRIX_NETSCALER_ACCESS_COMBINED_RESPONSE_TIME = LogFormat(
    "combined_resptime",
    re.compile(
        rf"""
        {RE_REMOTE_PATTERN}                  # remote_ip, remote_logname, remote_user
        \s
        {RE_ACCESS_COMMON_PATTERN}           # Timestamp, pid, method, uri, protocol, status code, bytes_sent
        \s
        {RE_REFERER_USER_AGENT_PATTERN}      # Referer, user_agent
        \s
        {RE_RESPONSE_TIME_PATTERN}           # Response time
        """,
        re.VERBOSE,
    ),
)

LOG_FORMAT_CITRIX_NETSCALER_ACCESS_COMBINED_RESPONSE_TIME_WITH_HEADERS = LogFormat(
    "combined_resptime_with_citrix_hdrs",
    re.compile(
        rf"""
        (?P<remote_ip>.*?)                   # Client IP address of the request.
        \s
        ->
        \s
        (?P<local_ip>.*?)                    # Local IP of the Netscaler.
        \s
        (?P<remote_logname>.*?)              # Remote logname (from identd, if supplied).
        \s
        (?P<remote_user>.*?)                 # Remote user if the request was authenticated.
        \s
        {RE_ACCESS_COMMON_PATTERN}           # Timestamp, pid, method, uri, protocol, status code, bytes_sent
        \s
        {RE_REFERER_USER_AGENT_PATTERN}      # Referer, user_agent
        \s
        {RE_RESPONSE_TIME_PATTERN}           # Response time
        """,
        re.VERBOSE,
    ),
)


class CitrixWebserverPlugin(ApachePlugin):
    """Apache log parsing plugin for Citrix specific logs.

    Citrix uses Apache with custom access log formats. These are::

        LogFormat "%{Citrix-ns-orig-srcip}i -> %{Citrix-ns-orig-destip}i %l %u %t [%P] \"%r\" %>s %b \"%{Referer}i\" \"%{User-Agent}i\" \"Time: %D microsecs\"" combined_resptime_with_citrix_hdrs
        LogFormat "%a %l %u %t [%P] \"%r\" %>s %b \"%{Referer}i\" \"%{User-Agent}i\" \"Time: %D microsecs\"" combined_resptime
    """  # noqa: E501

    __namespace__ = "citrix"

    ACCESS_LOG_NAMES = (*ApachePlugin.ACCESS_LOG_NAMES, "httpaccess.log", "httpaccess-vpn.log")
    ERROR_LOG_NAMES = (*ApachePlugin.ERROR_LOG_NAMES, "httperror.log", "httperror-vpn.log")

    def check_compatible(self) -> None:
        if not self.target.os == OperatingSystem.CITRIX:
            raise UnsupportedPluginError("Target is not a Citrix Netscaler")

    @staticmethod
    def infer_access_log_format(line: str) -> LogFormat:
        splitted_line = line.split()
        second_part = splitted_line[1]
        if second_part == "->":
            return LOG_FORMAT_CITRIX_NETSCALER_ACCESS_COMBINED_RESPONSE_TIME_WITH_HEADERS
        if "Time: " in line:
            return LOG_FORMAT_CITRIX_NETSCALER_ACCESS_COMBINED_RESPONSE_TIME

        return ApachePlugin.infer_access_log_format(line)



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.hypervisor import vmx

from dissect.target import container
from dissect.target.loader import Loader

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.target import Target


class VmxLoader(Loader):
    """Load VMware virtual machine configuration (VMX) files.

    References:
        - https://docs.vmware.com/en/VMware-Workstation-Pro/17/com.vmware.ws.using.doc/GUID-A968EF50-BA25-450A-9D1F-F8A9DEE640E7.html
    """

    def __init__(self, path: Path, **kwargs):
        super().__init__(path, **kwargs)
        self.vmx = vmx.VMX.parse(path.read_text())

    @staticmethod
    def detect(path: Path) -> bool:
        return path.suffix.lower() in (".vmx", ".vmtx")

    def map(self, target: Target) -> None:
        for disk in self.vmx.disks():
            path = self.base_path.joinpath(disk)

            if not path.is_file():
                base, sep, snapshot_id = path.stem.rpartition("-")
                if sep and len(snapshot_id) == 6 and snapshot_id.isdigit():
                    # Probably a snapshot, try to load the parent disk
                    target.log.info(
                        "Disk not found but seems to be a snapshot, trying previous snapshots: %s", path.name
                    )

                    snapshot_num = int(snapshot_id)
                    missing = [path.name]
                    for i in range(snapshot_num - 1, -1, -1):
                        snapshot_disk = (
                            path.with_name(f"{base}{path.suffix}")
                            if i == 0
                            else path.with_name(f"{base}-{i:06d}{path.suffix}")
                        )
                        target.log.debug("Trying to load snapshot: %s", snapshot_disk.name)

                        if snapshot_disk.is_file():
                            target.log.warning(
                                "Missing disk(s) but continuing with found snapshot: %s (missing %s)",
                                snapshot_disk.name,
                                ", ".join(missing),
                            )
                            path = snapshot_disk
                            break

                        missing.append(snapshot_disk.name)
                    else:
                        target.log.error("Failed to find previous snapshot for disk: %s", path.name)
                        continue
                else:
                    target.log.error("Disk not found: %s", path.name)
                    continue

            try:
                target.disks.add(container.open(path))
            except Exception:
                target.log.exception("Failed to load VMDK: %s", disk)



from __future__ import annotations

from typing import TYPE_CHECKING, BinaryIO

import dissect.cramfs as cramfs
from dissect.cramfs.c_cramfs import c_cramfs

from dissect.target.exceptions import (
    FileNotFoundError,
    FilesystemError,
    IsADirectoryError,
    NotADirectoryError,
    NotASymlinkError,
)
from dissect.target.filesystem import DirEntry, Filesystem, FilesystemEntry
from dissect.target.helpers import fsutil

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.cramfs import INode


class CramfsFilesystem(Filesystem):
    __type__ = "cramfs"

    def __init__(self, fh: BinaryIO, *args, **kwargs):
        super().__init__(fh, *args, **kwargs)
        self.cramfs = cramfs.CramFS(fh)

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        """Detect a CramFS filesystem on a given file-like object."""
        return int.from_bytes(fh.read(4), "little") == c_cramfs.CRAMFS_MAGIC

    def get(self, path: str) -> FilesystemEntry:
        return CramfsFilesystemEntry(self, path, self._get_node(path))

    def _get_node(self, path: str, node: INode | None = None) -> INode:
        """Returns an internal CramFS inode for a given path and optional relative inode."""
        try:
            return self.cramfs.get(path, node)
        except cramfs.FileNotFoundError as e:
            raise FileNotFoundError(path) from e
        except cramfs.NotADirectoryError as e:
            raise NotADirectoryError(path) from e
        except cramfs.NotASymlinkError as e:
            raise NotASymlinkError(path) from e
        except cramfs.Error as e:
            raise FileNotFoundError(path) from e


class CramfsDirEntry(DirEntry):
    fs: CramfsFilesystem
    entry: INode

    def get(self) -> CramfsFilesystemEntry:
        return CramfsFilesystemEntry(self.fs, self.path, self.entry)

    def stat(self, *, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self.get().stat(follow_symlinks=follow_symlinks)


class CramfsFilesystemEntry(FilesystemEntry):
    fs: CramfsFilesystem
    entry: INode

    def get(self, path: str) -> FilesystemEntry:
        full_path = fsutil.join(self.path, path, alt_separator=self.fs.alt_separator)
        return CramfsFilesystemEntry(self.fs, full_path, self.fs._get_node(path, self.entry))

    def open(self) -> BinaryIO:
        """Returns file handle (file-like object)."""
        if self.is_dir():
            raise IsADirectoryError(self.path)
        return self._resolve().entry.open()

    def scandir(self) -> Iterator[CramfsDirEntry]:
        """List the directory contents of this directory. Returns a generator of filesystem entries."""
        if not self.is_dir():
            raise NotADirectoryError(self.path)

        for entry in self._resolve().entry.iterdir():
            if entry.name in (".", ".."):
                continue

            yield CramfsDirEntry(self.fs, self.path, entry.name, entry)

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        """Return whether this entry is a directory."""
        try:
            return self._resolve(follow_symlinks=follow_symlinks).entry.is_dir()
        except FilesystemError:
            return False

    def is_file(self, follow_symlinks: bool = True) -> bool:
        """Return whether this entry is a file."""
        try:
            return self._resolve(follow_symlinks=follow_symlinks).entry.is_file()
        except FilesystemError:
            return False

    def is_symlink(self) -> bool:
        """Return whether this entry is a link."""
        return self.entry.is_symlink()

    def readlink(self) -> str:
        """Read the link of the given path if it is a symlink. Returns a string."""
        if not self.is_symlink():
            raise NotASymlinkError(self.path)

        return self.entry.link

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        """Return the stat information of this entry."""
        return self._resolve(follow_symlinks=follow_symlinks).lstat()

    def lstat(self) -> fsutil.stat_result:
        """Return the stat information of the given path, without resolving links."""
        node = self.entry

        # mode, ino, dev, nlink, uid, gid, size, ..., blocksize, numblocks
        st_info = [
            node.mode,
            node.offset,  # use inode offset as inode number
            id(self.fs),  # device ID of the filesystem
            1,  # cramfs inodes always have 1 nlinks
            node.uid,
            node.gid,
            node.size,
            0,  # cramfs inodes don't have timestamps
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            c_cramfs.CRAMFS_BLOCK_SIZE,
            len(node.blocks),
        ]

        return fsutil.stat_result(st_info)



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING

from dissect.target.helpers.regutil import VirtualHive, VirtualKey
from dissect.target.plugins.apps.productivity.sevenzip import SevenZipPlugin

if TYPE_CHECKING:
    from dissect.target.target import Target


def test_sevenzip_windows(target_win_users: Target, hive_hku: VirtualHive) -> None:
    """Test if we detect compression and extraction GUI dialog artifacts in 7-Zip."""

    # Compression dialog artifact
    key_path = "Software\\7-Zip\\Compression"
    key = VirtualKey(hive_hku, key_path)
    key.add_value(
        "ArcHistory",
        bytes.fromhex(
            "43003a005c00550073006500720073005c00410064006d0069006e006900730074007200610074006f007200"
            "5c0044006f00630075006d0065006e00740073005c006500780061006d0070006c0065002e0037007a000000"
        ),
    )
    key.timestamp = datetime(2025, 12, 31, 12, 34, 0, tzinfo=timezone.utc)
    hive_hku.map_key(key_path, key)

    # Extraction dialog artifact
    key_path = "Software\\7-Zip\\Extraction"
    key = VirtualKey(hive_hku, key_path)
    key.add_value(
        "PathHistory",
        bytes.fromhex(
            "43003a005c00550073006500720073005c00410064006d0069006e006900730074007200610074006f007200"
            "5c004400650073006b0074006f0070005c006500780061006d0070006c0065005c000000"
        ),
    )
    key.timestamp = datetime(2025, 12, 31, 13, 37, 0, tzinfo=timezone.utc)
    hive_hku.map_key(key_path, key)

    target_win_users.add_plugin(SevenZipPlugin)

    results = list(target_win_users.sevenzip())
    assert len(results) == 2

    assert results[0]._desc.name == "application/productivity/sevenzip/archistory"
    assert results[0].ts == datetime(2025, 12, 31, 12, 34, 0, tzinfo=timezone.utc)
    assert results[0].path == "C:\\Users\\Administrator\\Documents\\example.7z"

    assert results[1]._desc.name == "application/productivity/sevenzip/pathhistory"
    assert results[1].ts == datetime(2025, 12, 31, 13, 37, 0, tzinfo=timezone.utc)
    assert results[1].path == "C:\\Users\\Administrator\\Desktop\\example"



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

EnvironmentVariableRecord = TargetRecordDescriptor(
    "linux/proc/environ",
    [
        ("datetime", "ts"),
        ("string", "name"),
        ("varint", "pid"),
        ("string", "variable"),
        ("string", "content"),
    ],
)


class EnvironPlugin(Plugin):
    """Linux volatile proc environment plugin."""

    def check_compatible(self) -> None:
        if not self.target.has_function("proc"):
            raise UnsupportedPluginError("proc filesystem not available")

    @export(record=EnvironmentVariableRecord)
    def environ(self) -> Iterator[EnvironmentVariableRecord]:
        """Return the initial environment for all processes when they were started via execve(2).

        If the process modified its environment (e.g., by calling functions such as putenv(3) or modifying
        the environ(7) variable directly), this plugin will not reflect those changes.

        Yields EnvironmentVariableRecord with the following fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            ts (datetime): The modification timestamp of the processes' environ file.
            name (string): The name associated to the pid.
            pid (varint): The process id (pid) of the process.
            variable (string): The name of the environment variable.
            content (string): The contents of the environment variable.
        """
        for process in self.target.proc.processes():
            for environ in process.environ():
                yield EnvironmentVariableRecord(
                    ts=process.get("environ").stat().st_mtime,
                    name=process.name,
                    pid=process.pid,
                    variable=environ.variable,
                    content=environ.contents,
                    _target=self.target,
                )



from __future__ import annotations

from typing import TYPE_CHECKING

from flow.record.fieldtypes import datetime as dt

from dissect.target.helpers.regutil import VirtualHive, VirtualKey, VirtualValue
from dissect.target.plugins.os.windows.regf.muicache import MuiCachePlugin

if TYPE_CHECKING:
    from dissect.target.target import Target


def test_muicache_plugin(target_win_users: Target, hive_hku: VirtualHive) -> None:
    # NT >= 6.0
    muicache_shell_key = VirtualKey(
        hive_hku,
        "Software\\Classes\\Local Settings\\Software\\Microsoft\\Windows\\Shell\\MuiCache",
    )
    muicache_shell_key.add_value("LangID", VirtualValue(hive_hku, "LangID", b"\t\x04"))
    muicache_shell_key.add_value(
        "C:\\Windows\\System32\\fsquirt.exe.FriendlyAppName",
        VirtualValue(hive_hku, "C:\\Windows\\System32\\fsquirt.exe.FriendlyAppName", "fsquir"),
    )
    muicache_shell_key.add_value(
        "C:\\Windows\\System32\\fsquirt.test.exe.ApplicationCompany",
        VirtualValue(hive_hku, "C:\\Windows\\System32\\fsquirt.test.exe.ApplicationCompany", "fsquir"),
    )
    muicache_shell_key._timestamp = dt("2025-01-01 10:00:00.123456+00:00")

    # NT >= 6.0 (subkey)
    muicache_key = VirtualKey(
        hive_hku,
        "Software\\Classes\\Local Settings\\MuiCache\\35\\52C64B7E",
    )
    muicache_key.add_value(
        "@C:\\Windows\\system32\\wsecedit.dll,-718",
        VirtualValue(hive_hku, "@C:\\Windows\\system32\\wsecedit.dll,-718", "Local Security Policy"),
    )

    # NT < 6.0
    muicache_shellnoroam_key = VirtualKey(
        hive_hku,
        "Software\\Microsoft\\Windows\\ShellNoRoam\\MUICache",
    )
    muicache_shellnoroam_key.add_value(
        "@C:\\WINDOWS\\inf\\unregmp2.exe,-4",
        VirtualValue(hive_hku, "@C:\\WINDOWS\\inf\\unregmp2.exe,-4", "Windows Media Player"),
    )

    hive_hku.map_key(muicache_shell_key.path, muicache_shell_key)
    hive_hku.map_key(muicache_key.path, muicache_key)
    hive_hku.map_key(muicache_shellnoroam_key.path, muicache_shellnoroam_key)

    target_win_users.add_plugin(MuiCachePlugin)

    results = list(target_win_users.muicache())

    assert len(results) == 4
    assert results[0].regf_mtime == dt("2025-01-01 10:00:00.123456+00:00")
    assert results[0].index == 1
    assert results[0].name == "FriendlyAppName"
    assert results[0].value == "fsquir"
    assert str(results[0].path) == "C:\\Windows\\System32\\fsquirt.exe"
    assert results[1].index == 2
    assert results[1].name == "ApplicationCompany"
    assert results[1].value == "fsquir"
    assert str(results[1].path) == "C:\\Windows\\System32\\fsquirt.test.exe"
    assert results[2].index == 0
    assert results[2].name is None
    assert results[2].value == "Local Security Policy"
    assert str(results[2].path) == "@C:\\Windows\\system32\\wsecedit.dll"
    assert results[3].index == 0
    assert results[3].name is None
    assert results[3].value == "Windows Media Player"
    assert str(results[3].path) == "@C:\\WINDOWS\\inf\\unregmp2.exe"



from __future__ import annotations

from typing import TYPE_CHECKING

from defusedxml import ElementTree

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import ChildTargetRecord
from dissect.target.plugin import ChildTargetPlugin

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from dissect.target.helpers.fsutil import TargetPath
    from dissect.target.target import Target

PARALLELS_USER_PATHS = [
    "Parallels",
    "Documents/Parallels",
    "Library/Group Containers/*.com.parallels.desktop.appstore/Shared/Parallels",
]

PARALLELS_SYSTEM_PATHS = [
    "/Users/Shared/Parallels",
]


def find_pvms(target: Target) -> Iterator[TargetPath]:
    """Finds virtual machines located in default folders on a macOS target.

    References:
        - https://kb.parallels.com/117333
    """
    for user_details in target.user_details.all_with_home():
        for parallels_path in PARALLELS_SYSTEM_PATHS:
            if (path := target.fs.path(parallels_path)).exists():
                yield from iter_vms(path)

        for parallels_path in PARALLELS_USER_PATHS:
            if "*" in parallels_path:
                start_path, pattern = parallels_path.split("*", 1)
                for path in user_details.home_path.joinpath(start_path).rglob("*" + pattern):
                    yield from iter_vms(path)
            else:
                if (path := user_details.home_path.joinpath(parallels_path)).exists():
                    yield from iter_vms(path)


def iter_vms(path: Path) -> Iterator[TargetPath]:
    """Glob for .pvm folders in the provided folder."""
    for file in path.rglob("*.pvm"):
        if file.is_dir():
            yield file


class ParallelsChildTargetPlugin(ChildTargetPlugin):
    """Child target plugin that yields Parallels Desktop VM files."""

    __type__ = "parallels"

    def __init__(self, target: Target):
        super().__init__(target)
        self.pvms = list(find_pvms(target))

    def check_compatible(self) -> None:
        if not self.pvms:
            raise UnsupportedPluginError("No Parallels pvm file(s) found")

    def list_children(self) -> Iterator[ChildTargetRecord]:
        for pvm in self.pvms:
            try:
                config = ElementTree.fromstring(pvm.joinpath("config.pvs").read_bytes())
                name = config.find(".//VmName").text
            except Exception as e:
                self.target.log.error("Failed parsing VmName from config.pvs: %s", pvm)  # noqa: TRY400
                self.target.log.debug("", exc_info=e)

                name = None

            yield ChildTargetRecord(
                type=self.__type__,
                name=name,
                path=pvm,
                _target=self.target,
            )



from __future__ import annotations

import io
import re
from typing import TYPE_CHECKING

import pytest

from dissect.target.containers.raw import RawContainer
from dissect.target.plugins.scrape.qfind import QFindPlugin

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.plugins.scrape.qfind import QFindMatchRecord
    from dissect.target.target import Target

re_ansi_escape = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")


@pytest.fixture
def mock_target(target_bare: Target) -> Target:
    target_bare.add_plugin(QFindPlugin)

    buf = (
        (b"\x00" * 1024 * 64)
        + (b"ABCD" + b"\x00" * ((1024 * 4) - 4))
        + (b"A\x00B\x00C\x00D\x00" + b"\x00" * ((1024 * 4) - 8))
        + (b"\xab\xcd" + b"\x00" * ((1024 * 4) - 2))
        + (b"\x00" * 1024 * 52)
    )
    mock_disk = RawContainer(io.BytesIO(buf))
    target_bare.disks.add(mock_disk)

    return target_bare


def test_qfind(mock_target: Target) -> None:
    """Test if qfind can find a simple short needle."""

    results = list(mock_target.qfind(["B"]))

    assert len(results) == 2


def test_qfind_hex(mock_target: Target) -> None:
    """Test if qfind can find hex needles."""

    results: list[QFindMatchRecord] = list(mock_target.qfind(["ABCD"]))

    assert len(results) == 2

    assert results[0].disk == "<Container type=raw size=131072 vs=None>"
    assert results[0].offset == 0x10000
    assert results[0].needle == "ABCD"
    assert results[0].codec == "utf-8"

    assert results[1].disk == "<Container type=raw size=131072 vs=None>"
    assert results[1].offset == 0x12000
    assert results[1].needle == "ABCD"
    assert results[1].codec == "hex"


def test_qfind_needle_file(mock_target: Target, tmp_path: Path) -> None:
    """Test if qfind can work with a needle file."""

    needle_file = tmp_path.joinpath("needles.txt")
    needle_file.write_text("ABCD\n")

    results: list[QFindMatchRecord] = list(mock_target.qfind(needle_file=needle_file))

    assert len(results) == 2

    assert results[0].disk == "<Container type=raw size=131072 vs=None>"
    assert results[0].offset == 0x10000
    assert results[0].needle == "ABCD"

    assert results[1].disk == "<Container type=raw size=131072 vs=None>"
    assert results[1].offset == 0x12000
    assert results[1].needle == "ABCD"


def test_qfind_codecs(mock_target: Target) -> None:
    """Test if qfind can handle utf-8, utf-16-le and hex needles."""

    results: list[QFindMatchRecord] = list(mock_target.qfind(["ABCD"], encoding="utf-16-le"))
    assert len(results) == 3

    assert results[0].disk == "<Container type=raw size=131072 vs=None>"
    assert results[0].offset == 0x10000
    assert results[0].needle == "ABCD"
    assert results[0].codec == "utf-8"
    assert results[0].match == b"ABCD"

    assert results[1].offset == 0x11000
    assert results[1].needle == "ABCD"
    assert results[1].codec == "utf-16-le"
    assert results[1].match == b"ABCD"

    assert results[2].offset == 0x12000
    assert results[2].needle == "ABCD"
    assert results[2].codec == "hex"
    assert results[2].match == b"ABCD"

    results: list[QFindMatchRecord] = list(mock_target.qfind(["ABCD"], no_hex_decode=True))
    assert len(results) == 1

    assert results[0].disk == "<Container type=raw size=131072 vs=None>"
    assert results[0].offset == 0x10000
    assert results[0].needle == "ABCD"
    assert results[0].codec == "utf-8"
    assert results[0].match == b"ABCD"


def test_qfind_ignore_case(mock_target: Target) -> None:
    """Test if qfind can ignore case sensitivity of utf-8 needles."""

    results: list[QFindMatchRecord] = list(mock_target.qfind(["abcd"], ignore_case=True))

    assert len(results) == 2

    assert results[0].disk == "<Container type=raw size=131072 vs=None>"
    assert results[0].offset == 0x10000
    assert results[0].codec == "utf-8"
    assert results[0].needle == "abcd"
    assert results[0].match == b"ABCD"

    assert results[1].offset == 0x12000
    assert results[1].codec == "hex"
    assert results[1].needle == "abcd"
    assert results[1].match == b"\xab\xcd"


def test_qfind_regex(mock_target: Target) -> None:
    """Test if qfind can compile and search for regex needles."""

    results: list[QFindMatchRecord] = list(mock_target.qfind([r"[a-d]{4}"], regex=True, ignore_case=True))

    assert len(results) == 1
    assert results[0].offset == 0x10000
    assert results[0].needle == r"[a-d]{4}"
    assert results[0].codec == "utf-8"
    assert results[0].match == b"ABCD"
    assert results[0].buffer == (b"\x00" * 256) + b"ABCD" + (b"\x00" * 252)


def test_qfind_unique(target_bare: Target, capsys: pytest.CaptureFixture) -> None:
    """Test if qfind returns unique records when using the ``unique`` argument."""

    target_bare.add_plugin(QFindPlugin)

    buf = (b"\x00" * 512) + (b"ABCD" + b"\x00" * ((1024 * 8) - 4)) + (b"ABCD" + b"\x00" * ((1024 * 8) - 4))
    mock_disk = RawContainer(io.BytesIO(buf))
    target_bare.disks.add(mock_disk)

    results = list(target_bare.qfind(["ABCD"]))
    assert len(results) == 2

    results = list(target_bare.qfind(["ABCD"], unique=True))
    assert len(results) == 1



from __future__ import annotations

import ipaddress
import re
from collections import defaultdict
from typing import TYPE_CHECKING

from dissect.database.sqlite3 import SQLite3
from dissect.util.ts import from_unix

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

McAfeeMscLogRecord = TargetRecordDescriptor(
    "application/av/mcafee/msc/log",
    [
        ("datetime", "ts"),
        ("string", "threat"),
        ("string", "message"),
        ("string", "keywords"),
        ("string", "fkey"),
    ],
)

McAfeeMscFirewallRecord = TargetRecordDescriptor(
    "application/av/mcafee/msc/firewall",
    [
        ("datetime", "ts"),
        ("net.ipaddress", "ip"),
        ("uint16", "port"),
        ("string", "protocol"),
        ("string", "message"),
        ("string", "keywords"),
        ("string", "fkey"),
    ],
)

re_cdata = re.compile(r"<!\[CDATA\[(.*?)\]\]>", flags=re.MULTILINE)
re_strip_tags = re.compile(r"<[^!][^>]*>")


class McAfeePlugin(Plugin):
    """McAfee antivirus plugin."""

    __namespace__ = "mcafee"

    DIRS = (
        "sysvol/ProgramData/McAfee/MSC/Logs",  # Windows
        "/opt/McAfee/ens/log/tp",  # Linux/Mac according to docs
        "/opt/McAfee/ens/log/esp",  # Linux/Mac according to docs
    )
    LOG_FILE_PATTERN = "*.log"
    TEMPLATE_ID_INFECTION = 102
    MARKER_INFECTION = "%INFECTION_INFO%"
    MARKER_SUSPICIOUS_TCP_CONNECTION = "TCP port "
    MARKER_SUSPICIOUS_UDP_CONNECTION = "UDP port "
    TABLE_LOG = "log"
    TABLE_FIELD = "field"

    def check_compatible(self) -> None:
        if not list(self.get_log_files()):
            raise UnsupportedPluginError("No McAfee Log files found")

    def get_log_files(self) -> Iterator[Path]:
        for path in self.DIRS:
            yield from self.target.fs.path(path).glob(self.LOG_FILE_PATTERN)

    def _clean_message(self, message: str) -> str:
        return re.sub(re_strip_tags, "", (" ".join(re.findall(re_cdata, message))))

    @export(record=McAfeeMscLogRecord)
    def msc(self) -> Iterator[McAfeeMscLogRecord]:
        """Return msc log history records from McAfee.

        Yields McAfeeMscLogRecord with the following fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            ts (datetime): timestamp.
            ip (net.ipadress): IP of suspicious connection (if available).
            tcp_port (net.tcp.Port): TCP Port of suspicious incoming connection (if available).
            udp_port (net.udp.Port): UDP Port of suspicious incoming connection (if available).
            threat (string): Description of the detected threat (if available).
            message (string): Message as reported in the user interface (might include template slots).
            keywords (string): Unparsed fields that might be visible in user interface.
            fkey (string): Foreign key for reference for further investigation.
        """

        len_marker = len(self.MARKER_SUSPICIOUS_UDP_CONNECTION)

        for log_file in self.get_log_files():
            with SQLite3(log_file) as database:
                fields = defaultdict(dict)
                fields_table = database.table(self.TABLE_FIELD)

                for field in fields_table.rows():
                    fields[field.fkey][field.field_id] = field.data
                log_table = database.table(self.TABLE_LOG)

                for entry in log_table.rows():
                    fkey = entry.fkey
                    log_fields = fields[fkey]
                    ip = None
                    protocol = None
                    port = None
                    threat = None

                    for key, log_field in log_fields.items():
                        try:
                            ipaddress.ip_address(log_field)
                            ip = log_field
                            continue
                        except ValueError:
                            pass

                        if log_field.startswith(
                            (self.MARKER_SUSPICIOUS_TCP_CONNECTION, self.MARKER_SUSPICIOUS_UDP_CONNECTION)
                        ):
                            port = int(log_field[len_marker:])
                            protocol = log_field[:3]
                            continue

                        if key == self.TEMPLATE_ID_INFECTION and entry.details_info.find(self.MARKER_INFECTION) > -1:
                            threat = log_field

                    if threat:
                        yield McAfeeMscLogRecord(
                            ts=from_unix(entry.date),
                            threat=threat,
                            message=self._clean_message(entry.details_info),
                            keywords=",".join(log_fields.values()),
                            fkey=entry.fkey,
                        )
                    else:
                        yield McAfeeMscFirewallRecord(
                            ts=from_unix(entry.date),
                            ip=ip,
                            protocol=protocol,
                            port=port,
                            message=self._clean_message(entry.details_info),
                            keywords=",".join(log_fields.values()),
                            fkey=entry.fkey,
                        )



from __future__ import annotations

import io
from collections import defaultdict
from typing import TYPE_CHECKING, BinaryIO

from dissect.util.stream import MappingStream

from dissect.target.helpers.scrape import Needle, find_needles, scrape_chunks
from dissect.target.plugin import Plugin, internal
from dissect.target.volume import EncryptedVolumeSystem, LogicalVolumeSystem, Volume

if TYPE_CHECKING:
    import re
    from collections.abc import Callable, Iterator

    from dissect.target.container import Container
    from dissect.target.helpers.record import TargetRecordDescriptor


class ScrapePlugin(Plugin):
    __namespace__ = "scrape"

    def check_compatible(self) -> None:
        pass

    def create_streams(
        self, *, encrypted: bool = True, lvm: bool = True, all: bool = False
    ) -> Iterator[tuple[Container | Volume, MappingStream]]:
        """Yields streams for all disks and volumes of a target.

        At the basis, each disk of a target is represented as a stream of itself. If the target has volumes, these are
        checked to be either encrypted or LVM volumes (if the respective flags are set). If so, that volume is replaced
        in the disk stream with a stream of the volume itself, or removed.

        In the case of encrypted volumes, the encrypted volume is replaced with the decrypted volume. This is done
        transparently on the disk stream level, so the consumer does not need to worry about the decryption process.

        In the case of LVM volumes (including any sort of RAID), the base volumes are removed from the disk stream,
        and a new stream will be yielded for each logical volume. This ensures that no unnecessary scraping is
        performed on base volumes of a logical volume, but only on the reconstructed logical volume.

        Args:
            encrypted: Whether to replace encrypted volumes for their decrypted counterpart.
            lvm: Whether to skip base volumes of LVM volumes and yield the logical volume itself.
            all: Whether to yield all disks and volumes, regardless of their type.

        Yields:
            A tuple containing the disk (or volume, in the case of an LVM volume) and a stream of that disk.
        """
        # Create a map of disk regions we want to scrape
        scrape_map = defaultdict(dict)

        # Start off by making a very simple `disk: {(offset, size): volume}` map for every volume
        # Spaces in between volumes are also added as disk regions
        offset = 0
        for disk in self.target.disks:
            for volume in disk.vs.volumes if disk.vs else []:
                if offset != volume.offset:
                    # There's data between the volumes, so add that from the disk
                    scrape_map[disk][(offset, volume.offset - offset)] = (disk, offset)
                    offset = volume.offset

                # Add the volume to the stream
                scrape_map[disk][(volume.offset, volume.size)] = (volume, 0)
                offset += volume.size

            # There's data after the volumes until the end of the disk
            # Coincidentally this also takes care of disks with no volumes
            if offset != disk.size:
                scrape_map[disk][(offset, disk.size - offset)] = (disk, offset)

        # Iterate over all volumes and add the decrypted variant of encrypted volumes
        for volume in self.target.volumes:
            if encrypted and isinstance(volume.vs, EncryptedVolumeSystem):
                if not all:
                    # Decrypted volumes have the encrypted volume as the disk
                    source_volume: Volume = volume.disk
                    source_disk = source_volume.disk

                    # Replace the encrypted volume region with the decrypted volume region
                    scrape_map[source_disk][(source_volume.offset, source_volume.size)] = (volume, 0)
                else:
                    # Add the encrypted volume separately to the map
                    scrape_map[volume][(0, volume.size)] = (volume, 0)

            if lvm and isinstance(volume.vs, LogicalVolumeSystem):
                if not all:
                    # Remove the base volumes from the map
                    for source_volume in volume.disk:
                        source_disk = source_volume.disk
                        del scrape_map[source_disk][(source_volume.offset, source_volume.size)]

                # Add the logical volumes to the map
                scrape_map[volume][(0, volume.size)] = (volume, 0)

        for disk, volumes in scrape_map.items():
            # If we ended up removing all regions from the disk, skip it
            # I.e. when a disk is fully occupied by an LVM volume
            if not volumes:
                continue

            size = 0
            stream = MappingStream()
            for (map_offset, map_size), (source, source_offset) in volumes.items():
                stream.add(map_offset, map_size, source, source_offset)
                size = max(size, map_offset + map_size)
            stream.size = size

            yield disk, stream

    @internal
    def find(
        self,
        needles: Needle | list[Needle],
        lock_seek: bool = True,
        block_size: int = io.DEFAULT_BUFFER_SIZE,
        progress: Callable[[Container | Volume, int, int], None] | None = None,
    ) -> Iterator[tuple[Container | Volume, MappingStream, Needle, int, re.Match | None]]:
        """Yields needles, their offsets and an optional regex match found in all disks and volumes of a target.

        Args:
            needles: The needle or list of needles to search for.
            lock_seek:  Whether the file position is maintained by the scraper or the consumer.
                        Setting this to ``False`` wil allow the consumer to seek the file pointer, i.e. to skip forward.
            block_size: The block size to use for reading from the byte stream.
            progress: A function to call with the current disk, offset and size of the stream.
        """
        for disk, stream in self.create_streams():
            for needle, offset, match in find_needles(
                stream,
                needles,
                lock_seek=lock_seek,
                block_size=block_size,
                progress=(lambda current, disk=disk, stream=stream: progress(disk, current, stream.size))
                if progress
                else None,
            ):
                yield disk, stream, needle, offset, match

    @internal
    def scrape_chunks_from_disks(
        self,
        chunk_parser: Callable[[Needle, bytes], Iterator[TargetRecordDescriptor]],
        needle_chunk_size_map: dict[Needle, int] | None = None,
        needle: Needle | None = None,
        chunk_size: int | None = None,
        chunk_reader: Callable[[BinaryIO, Needle, int, int], bytes] | None = None,
        block_size: int = io.DEFAULT_BUFFER_SIZE,
    ) -> Iterator[TargetRecordDescriptor]:
        """Yields records scraped from chunks found in ``target.disks``.

        Args:
            chunk_parser: A function to parse a chunk and yield records.
            needle_chunk_size_map: A dictionary with needle as keys and chunk sizes as values.
            needle: A single needle to search for.
            chunk_size: The size of the chunks to scrape, when providing a single needle.
            chunk_reader: A function to read a chunk from a byte stream for provided needle, offset and chunk size.
            block_size: The block size to use for reading from the byte stream.
        """

        if needle_chunk_size_map and (needle or chunk_size):
            raise ValueError("Either `needle_chunk_size_map` or both `needle` and `chunk_size` must be provided")

        if (needle and not chunk_size) or (not needle and chunk_size):
            raise ValueError("Both `needle` and `chunk_size` must be provided")
        else:
            needle_chunk_size_map = {needle: chunk_size}

        for disk in self.target.disks:
            disk.seek(0)
            yield from scrape_chunks(
                disk,
                needle_chunk_size_map=needle_chunk_size_map,
                chunk_parser=chunk_parser,
                chunk_reader=chunk_reader,
                block_size=block_size,
            )

    @internal
    def scrape_needles_from_disks(
        self,
        needle: Needle | None = None,
        needles: list[Needle] | None = None,
        block_size: int = io.DEFAULT_BUFFER_SIZE,
    ) -> Iterator[tuple[BinaryIO, bytes, int, re.Match | None]]:
        """Yields ``(bytestream, needle, offset)`` tuples, scraped from ``target.disks``.

        Args:
            needles: A list of byte needles to search for.
            needle: A single byte needle to search for.
            block_size: The block size to use for reading from the byte stream.
        """

        if needle and needles:
            raise ValueError("Either `needles` values or a single `needle` value must be provided")
        elif not needle and not needles:
            raise ValueError("At least one needle value must be provided")

        if needle:
            needles = [needle]

        for disk in self.target.disks:
            disk.seek(0)
            for needle, offset, match in find_needles(disk, needles=needles, block_size=block_size):
                yield disk, needle, offset, match



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import patch

import pytest

from dissect.target.loader import open as loader_open
from dissect.target.loaders.tar import TarLoader
from dissect.target.loaders.vmsupport import VmSupportLoader, VmSupportTarSubloader
from dissect.target.target import Target
from tests._utils import absolute_path

if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path

    from pytest_benchmark.fixture import BenchmarkFixture

    from dissect.target.loader import Loader


@pytest.fixture
def mock_vmsupport_dir(tmp_path: Path) -> Path:
    root = tmp_path / "esx-localhost-2026-01-09--16.04-135806"

    (root / "etc" / "vmware").mkdir(parents=True)
    (root / "action.log").write_bytes(b"")
    (root / "error.log").write_bytes(b"")
    (root / "etc" / "vmware" / "esx.conf").write_bytes(b'/resourceGroups/version = "8.0.3"\n')
    return tmp_path


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
# vmsupport files with commands/vsi_traverse_-s* removed (~300mo)
@pytest.mark.parametrize(
    ("path", "loader"),
    [
        ("_data/loaders/vmsupport/esx-localhost6-2026-01-12--13.56-2107676.tar.gz", TarLoader),
        ("_data/loaders/vmsupport/esx-localhost8-2026-01-09--16.04-135806.tgz", TarLoader),
        ("mock_vmsupport_dir", VmSupportLoader),
    ],
)
def test_target_open(
    opener: Callable[[str | Path], Target], path: str, loader: type[Loader], mock_vmsupport_dir: Path
) -> None:
    """Test that we correctly use the ESXi vm-support loaders when opening a ``Target``."""
    path = mock_vmsupport_dir if path == "mock_vmsupport_dir" else absolute_path(path)

    with patch("dissect.target.target.Target.apply"):
        target = opener(path)

        assert isinstance(target._loader, loader)
        if isinstance(target._loader, TarLoader):
            assert isinstance(target._loader.subloader, VmSupportTarSubloader)
        assert target.path == path


@pytest.mark.parametrize(
    "data_path",
    [
        "_data/loaders/vmsupport/esx-localhost6-2026-01-12--13.56-2107676.tar.gz",
        "_data/loaders/vmsupport/esx-localhost8-2026-01-09--16.04-135806.tgz",
    ],
)
def test_compressed_tar(data_path: str) -> None:
    """Test if we map a compressed vm support tar image correctly."""
    path = absolute_path(data_path)

    loader = loader_open(path)
    assert isinstance(loader, TarLoader)

    t = Target()
    loader.map(t)
    assert isinstance(loader.subloader, VmSupportTarSubloader)
    assert len(t.filesystems) == 1

    t.apply()
    test_file = t.fs.path("/etc/vmware/esx.conf")
    assert test_file.exists()
    assert test_file.is_file()
    assert b"/resourceGroups/version" in test_file.open().read()


def test_dir(mock_vmsupport_dir: Path) -> None:
    """Test if we map an extracted vm support directory correctly."""

    loader = loader_open(mock_vmsupport_dir)
    assert isinstance(loader, VmSupportLoader)

    t = Target()
    loader.map(t)
    assert len(t.filesystems) == 1

    t.apply()
    test_file = t.fs.path("etc/vmware/esx.conf")
    assert test_file.exists()
    assert test_file.is_file()
    assert test_file.open().readline() == b'/resourceGroups/version = "8.0.3"\n'


@pytest.mark.parametrize(
    ("archive", "loader"),
    [
        ("_data/loaders/vmsupport/esx-localhost6-2026-01-12--13.56-2107676.tar.gz", TarLoader),
        ("_data/loaders/vmsupport/esx-localhost8-2026-01-09--16.04-135806.tgz", TarLoader),
    ],
)
@pytest.mark.benchmark
def test_benchmark(benchmark: BenchmarkFixture, archive: str, loader: type[Loader]) -> None:
    file = absolute_path(archive)

    benchmark(lambda: loader(file).map(Target()))



from collections.abc import Iterator
from typing import ClassVar

from dissect.target import Target
from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.fsutil import TargetPath
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export
from dissect.target.plugins.os.windows.everything.parser import (
    EverythingDB,
)

EverythingRecord = TargetRecordDescriptor(
    "windows/everything/record",
    [
        ("path", "path"),
        ("filesize", "size"),
        ("datetime", "date_created"),
        ("datetime", "date_modified"),
        ("datetime", "date_accessed"),
        ("uint32", "attributes"),
        ("string", "type"),
        ("string", "source"),
    ],
)


class EverythingPlugin(Plugin):
    """Voidtools Everything database parser.

    Extracts files and metadata from the Everything database, which is stored in a proprietary format.
    """

    __namespace__ = "everything"

    PATH_GLOBS: ClassVar = [
        "sysvol\\Program Files\\Everything\\Everything*.db",
        "sysvol\\Program Files (x86)\\Everything\\Everything*.db",
    ]
    USER_PATH: ClassVar = "AppData\\Local\\Everything\\Everything*.db"

    def __init__(self, target: Target):
        super().__init__(target)
        self.configs = []
        for path_option in self.PATH_GLOBS:
            self.configs.extend(self.target.fs.path().glob(path_option))

        self.configs.extend(self.find_user_files())

    def find_user_files(self) -> Iterator[TargetPath]:
        for user_details in self.target.user_details.all_with_home():
            yield from user_details.home_path.glob(self.USER_PATH)

    def check_compatible(self) -> None:
        if not self.configs:
            raise UnsupportedPluginError("No everything.db files found")

    @export(record=EverythingRecord)
    def locate(self) -> Iterator[EverythingRecord]:
        """Yield file and directory names from everything.db file."""
        for path in self.configs:
            try:
                with self.target.fs.path(path).open() as fh:
                    db = EverythingDB(fh)
                    for item in db:
                        yield EverythingRecord(
                            path=item.file_path,
                            size=item.size,
                            date_created=item.date_created,
                            date_modified=item.date_modified,
                            date_accessed=item.date_accessed,
                            attributes=item.attributes,
                            type=item.file_type,
                            source=path,
                            _target=self.target,
                        )
            except (NotImplementedError, ValueError) as e:  # noqa: PERF203
                self.target.log.warning("Invalid EverythingDB %s: %s", path, e)



from __future__ import annotations

import os
import string
import urllib.parse
from pathlib import Path
from typing import TYPE_CHECKING, BinaryIO

from dissect.target.exceptions import FileNotFoundError
from dissect.target.filesystems.ntfs import NtfsFilesystem
from dissect.target.helpers.logging import get_logger

if TYPE_CHECKING:
    from dissect.target.filesystem import Filesystem
    from dissect.target.target import Target


log = get_logger(__name__)


def add_virtual_ntfs_filesystem(
    target: Target,
    fs: Filesystem,
    boot_path: str = "$Boot",
    mft_path: str = "$MFT",
    usnjrnl_path: str = "$Extend/$Usnjrnl:$J",
    sds_path: str = "$Secure:$SDS",
) -> None:
    """Utility for creating an NtfsFilesystem with separate system files from another Filesystem, usually
    a DirectoryFilesystem or VirtualFilesystem.

    Args:
        target: The target to add the filesystem to.
        fs: The Filesystem to load the system files from.
        boot_path: Path to open the $Boot file from.
        mft_path: Path to open the $MFT file from.
        usnjrnl_path: Path to open the $Usnjrnl:$J file from.
        sds_path: Path to open the $Secure:$SDS file from.
    """
    fh_boot = _try_open(fs, boot_path)
    fh_mft = _try_open(fs, mft_path)
    fh_usnjrnl = _try_open(fs, usnjrnl_path)
    fh_sds = _try_open(fs, sds_path)

    if any([fh_boot, fh_mft]):
        ntfs = None

        try:
            ntfs = NtfsFilesystem(boot=fh_boot, mft=fh_mft, usnjrnl=fh_usnjrnl, sds=fh_sds)
        except Exception as e:
            if fh_boot:
                log.warning("Failed to load NTFS filesystem from %s, retrying without $Boot file", fs)
                log.debug("", exc_info=e)

                try:
                    # Try once more without the $Boot file
                    ntfs = NtfsFilesystem(mft=fh_mft, usnjrnl=fh_usnjrnl, sds=fh_sds)
                except Exception:
                    log.warning("Failed to load NTFS filesystem from %s without $Boot file, skipping", fs)
                    return

        # Only add it if we have a valid NTFS with an MFT
        if ntfs and ntfs.ntfs.mft:
            target.filesystems.add(ntfs)
            fs.ntfs = ntfs.ntfs
        else:
            log.warning("Opened NTFS filesystem from %s but could not find $MFT, skipping", fs)


def _try_open(fs: Filesystem, path: str) -> BinaryIO | None:
    paths = [path] if not isinstance(path, list) else path

    for path in paths:
        try:
            path = fs.get(path)
            if path.stat().st_size > 0:
                return path.open()
            log.warning("File is empty and will be skipped: %s", path)
        except FileNotFoundError:  # noqa: PERF203
            pass
    return None


def parse_path_uri(path: str | Path | None) -> tuple[Path | None, urllib.parse.ParseResult | None]:
    """Parse a path URI into its path component and a parsed path, if applicable.

    If the path is a string, it will be parsed as a URI. If it is a Path object, it will be returned as is.

    Args:
        path: String or ``Path`` describing the path of a target.

    Returns:
        A tuple containing a normalized path and the parsed path.
    """
    if path is None:
        return None, None

    if isinstance(path, os.PathLike):
        return path, None

    parsed_path = urllib.parse.urlparse(path)
    if not parsed_path.scheme or (len(parsed_path.scheme) == 1 and parsed_path.scheme in string.ascii_letters):
        return Path(path), None

    return Path(parsed_path.netloc + parsed_path.path).expanduser(), parsed_path



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime, timezone
from itertools import islice
from typing import TYPE_CHECKING, Any, Final

from dissect.target.exceptions import RegistryValueNotFoundError, UnsupportedPluginError
from dissect.target.helpers.network import IANAProtocol
from dissect.target.helpers.record import DynamicDescriptor, TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator
    from pathlib import Path

    from dissect.target.helpers.regutil import RegistryKey
    from dissect.target.target import Target


WindowsFirewallLogRecord = TargetRecordDescriptor(
    "windows/firewall/log",
    [
        ("datetime", "ts"),
        ("string", "action"),
        ("string", "protocol"),
        ("net.ipaddress", "src_ip"),
        ("net.ipaddress", "dst_ip"),
        ("varint", "src_port"),
        ("varint", "dst_port"),
        ("filesize", "size"),
        ("string", "tcpflags"),
        ("string", "tcpsyn"),
        ("string", "tcpack"),
        ("string", "tcpwin"),
        ("string", "icmptype"),
        ("string", "icmpcode"),
        ("string", "info"),
        ("string", "path"),
        ("path", "source"),
    ],
)


class WindowsFirewallPlugin(Plugin):
    """Windows Firewall plugin."""

    __namespace__ = "firewall"

    RULE_KEYS = (
        # Defaults
        "HKLM\\SYSTEM\\CurrentControlSet\\Services\\SharedAccess\\Defaults\\FirewallPolicy\\FirewallRules",
        # Parameters
        "HKLM\\SYSTEM\\CurrentControlSet\\Services\\SharedAccess\\Parameters\\FirewallPolicy\\FirewallRules",
        "HKLM\\SYSTEM\\CurrentControlSet\\Services\\SharedAccess\\Parameters\\FirewallPolicy\\Mdm\\FirewallRules",
        # Legacy
        "HKLM\\SOFTWARE\\Policies\\Microsoft\\WindowsFirewall\\FirewallRules",
        # Other
        "HKLM\\SYSTEM\\CurrentControlSet\\Services\\SharedAccess\\Parameters\\FirewallPolicy\\RestrictedServices\\AppIso\\FirewallRules",
        "HKLM\\SYSTEM\\CurrentControlSet\\Services\\SharedAccess\\Parameters\\FirewallPolicy\\RestrictedServices\\Configurable\\System",
        "HKLM\\SYSTEM\\CurrentControlSet\\Services\\SharedAccess\\Parameters\\FirewallPolicy\\RestrictedServices\\Static\\System",
    )

    LOGGING_KEYS = (
        # Defaults
        "HKLM\\SYSTEM\\CurrentControlSet\\Services\\SharedAccess\\Defaults\\FirewallPolicy\\PublicProfile\\Logging"
        "HKLM\\SYSTEM\\CurrentControlSet\\Services\\SharedAccess\\Defaults\\FirewallPolicy\\StandardProfile\\Logging"
        "HKLM\\SYSTEM\\CurrentControlSet\\Services\\SharedAccess\\Defaults\\FirewallPolicy\\DomainProfile\\Logging"
        # Parameters
        "HKLM\\SYSTEM\\CurrentControlSet\\Services\\SharedAccess\\Parameters\\FirewallPolicy\\PublicProfile\\Logging",
        "HKLM\\SYSTEM\\CurrentControlSet\\Services\\SharedAccess\\Parameters\\FirewallPolicy\\StandardProfile\\Logging",
        "HKLM\\SYSTEM\\CurrentControlSet\\Services\\SharedAccess\\Parameters\\FirewallPolicy\\DomainProfile\\Logging",
    )

    def __init__(self, target: Target):
        super().__init__(target)
        self.keys = list(self.find_rule_keys())
        self.log_paths = list(self.find_log_paths())

    def find_rule_keys(self) -> Iterator[RegistryKey]:
        for key in self.RULE_KEYS:
            yield from self.target.registry.keys(key)

    def find_log_paths(self) -> Iterator[Path]:
        seen = set()

        for path in self.target.fs.path("sysvol/Windows/System32/LogFiles/Firewall").glob("pfirewall*"):
            seen.add(path)
            yield path

        if self.target.has_function("registry"):
            for log_key in self.LOGGING_KEYS:
                for key in self.target.registry.keys(log_key):
                    try:
                        log_path = self.target.resolve(key.value("LogFilePath").value)
                        if log_path.is_file() and log_path not in seen:
                            seen.add(log_path)
                            yield log_path
                    except (ValueError, RegistryValueNotFoundError):  # noqa: PERF203
                        pass

    def check_compatible(self) -> None:
        if not self.keys and not self.log_paths:
            raise UnsupportedPluginError("No Windows Firewall registry keys or log files found on target")

    @export(record=DynamicDescriptor(["string", "boolean"]))
    def rules(self) -> Iterator[DynamicDescriptor]:
        """Return firewall rules saved in the Windows registry.

        For a Windows operating system, the Firewall rules are stored in the
        ``HKLM\\SYSTEM\\CurrentControlSet\\Services\\SharedAccess\\Parameters\\FirewallPolicy\\FirewallRules`` registry key.

        References:
            - https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-gpfas/2efe0b76-7b4a-41ff-9050-1023f8196d16

        Yields dynamic records with usually the following fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            key (string): The rule key name.
            version (string): The version field of the rule.
            action (string): The action of the rule.
            active (boolean): Whether the rule is active.
            dir (string): The direction of the rule.
            protocol (string): The specified IANA protocol (UDP, TCP, etc).
            lport (string): The listening port or range of the rule.
            rport (string): The receiving port or range the rule.
            profile (string): The Profile field of the rule.
            app (string): The App field of the rule.
            svc (string): The Svc of the rule.
            name (string): The Name of the rule.
            desc (string): The Desc of the rule.
            embed_ctxt (string): The EmbedCtxt of the rule.
        """  # noqa: E501

        FIELD_MAP: Final[dict[str, str]] = {
            "app": "path",
            "active": "boolean",
            # protocol int is translated to a string at runtime
            # lport and rport can be a string (e.g. port ranges or other text)
            "lport": "string[]",
            "rport": "string[]",
        }
        VALUE_MAP: Final[dict[str, Callable[[str], Any]]] = {"active": lambda val: val.upper() == "TRUE"}

        for reg in self.keys:
            for entry in reg.values():
                r = [
                    ("string", "key"),
                    ("string", "version"),
                ]
                data = {}

                pairs: list[str] = entry.value.split("|")
                data["version"] = pairs.pop(0)

                for kv in pairs:
                    key, _, value = kv.partition("=")

                    if not (key := key.lower()):
                        continue

                    field_type = FIELD_MAP.get(key, "string")
                    r.append((field_type, key))

                    try:
                        value = VALUE_MAP[key](value)
                    except KeyError:
                        pass
                    data[key] = value

                # Resolve file paths to TargetPaths
                if "app" in data:
                    data["app"] = self.target.resolve(data["app"])

                # Translate protocol numbers to IANA names
                if protocol := data.get("protocol"):
                    if protocol.isnumeric():
                        data["protocol"] = IANAProtocol(int(protocol)).name
                    elif protocol.isalpha():
                        data["protocol"] = protocol.upper()

                # Make sure actions and directions are consistent
                if "action" in data:
                    data["action"] = data["action"].upper()
                if "dir" in data:
                    data["dir"] = data["dir"].upper()

                # Put port ranges in a list
                if "lport" in data:
                    data["lport"] = data["lport"].split(",")
                if "rport" in data:
                    data["rport"] = data["rport"].split(",")

                r.append(("path", "source"))

                yield TargetRecordDescriptor("windows/firewall/rule", r)(
                    key=entry.name,
                    **data,
                    source=f"HKLM\\{reg.path}",
                    _target=self.target,
                )

    @export(record=WindowsFirewallLogRecord)
    def logs(self) -> Iterator[WindowsFirewallLogRecord]:
        """Parse Windows Firewall log files.

        Currently parses ``pfirewall*`` files in ``sysvol\\Windows\\System32\\LogFiles\\Firewall\\`` only.
        Does not yet parse dynamically set log locations e.g. ``netsh advfirewall set currentprofile logging filename``.

        References:
            - https://learn.microsoft.com/en-us/windows/security/operating-system-security/network-security/windows-firewall/configure-logging

        Yields Windows Firewall log records with the following fields:

        .. code-block:: text

            ts (datetime): The timestamp of the log entry.
            hostname (string): The target hostname.
            domain (string): The target domain.
            action (string): Allow or Block.
            protocol (string): TCP, UDP or other IANA protocol value.
            src_ip (net.ipaddress): Source IP address.
            dst_ip (net.ipaddress): Destination IP address.
            src_port (varint): Source port number.
            dst_port (varint): Destination port number.
            size (filesize): Size in bytes of the packet(s).
            tcpflags (string): TCP header control flags.
            tcpsyn (string): TCP sequence number.
            tcpack (string): TCP acknowledgement number.
            tcpwin (string): TCP window size in bytes.
            icmptype (string): ICMP packet type.
            icmpcode (string): ICMP packet code.
            info (string): Additional information.
            path (string): Direction of the traffic, either SEND, RECEIVE, FORWARD or UNKNOWN.
            source (path): Source path of the record log line.
        """

        for path in self.log_paths:
            with path.open("rt") as fh:
                try:
                    header = list(islice(fh, 4))
                    config = LogConfig(header)
                except Exception as e:
                    self.target.log.warning("Unable to read Windows Firewall log file config in %s: %s", path, e)
                    self.target.log.debug("", exc_info=e)
                    continue

                if config.version != 1.5:
                    self.target.log.warning(
                        "Unsupported Windows Firewall log file version %r in log file %s", header[0], path
                    )
                    continue

                tzinfo = self.target.datetime.tzinfo if config.time_format == "local" else timezone.utc

                for line in fh:
                    if not (line := line.strip()):
                        continue

                    entry = {
                        k.replace("-", "_"): v if v != "-" else None
                        for k, v in zip(config.fields, line.split(" "), strict=False)
                    }

                    entry["ts"] = datetime.strptime(f"{entry['date']} {entry['time']}", "%Y-%m-%d %H:%M:%S").replace(
                        tzinfo=tzinfo
                    )
                    entry.pop("date")
                    entry.pop("time")

                    yield WindowsFirewallLogRecord(
                        **entry,
                        source=path,
                        _target=self.target,
                    )


@dataclass
class LogConfig:
    """Windows Firewall Log ``pfirewall.log`` file config parser.

    Fields can differ depending on configuration and version.

    Example structure of a regular ``pfirewall.log`` file:

    .. code-block::

        #Version: 1.5
        #Software: Microsoft Windows Firewall
        #Time Format: Local
        #Fields: date time action protocol src-ip dst-ip src-port dst-port size tcpflags tcpsyn tcpack tcpwin icmptype icmpcode info path

        2022-01-01 13:37:00 DROP UDP 1.2.3.4 5.6.7.8 1234 5678 1337 - - - - - - - RECEIVE
    """  # noqa: E501

    raw: list[str] = field(repr=False)
    version: float | None = None
    software: str | None = None
    time_format: str | None = None
    fields: list[str] | None = None

    def __post_init__(self):
        assert len(self.raw) == 4
        _, _, self.version = self.raw[0].strip().partition("#Version: ")
        _, _, self.software = self.raw[1].strip().partition("#Software: ")
        _, _, self.time_format = self.raw[2].strip().lower().partition("#time format: ")
        _, _, fields = self.raw[3].strip().partition("#Fields: ")
        self.version = float(self.version)
        self.fields = fields.split(" ")



from __future__ import annotations

from typing import TYPE_CHECKING

from flow.record.fieldtypes import datetime as dt

from dissect.target.plugins.apps.av.mcafee import McAfeeMscFirewallRecord, McAfeePlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_mcafee_plugin_log(target_win: Target, fs_win: VirtualFilesystem) -> None:
    log_dir = absolute_path("_data/plugins/apps/av/mcafee")
    fs_win.map_dir("ProgramData/McAfee/MSC/Logs", log_dir)

    target_win.add_plugin(McAfeePlugin)

    records = list(target_win.mcafee.msc())
    assert len(records) == 2
    for record in records:
        if isinstance(record, type(McAfeeMscFirewallRecord())):
            assert record.ip == "127.0.0.1"
            assert record.protocol == "TCP"
            assert record.port == 54996
            assert record.ts == dt("2023-03-07T10:32:34Z")
            assert record.fkey == "{C492A216-EFFC-4DAE-BE5E-2F5E064594C9}"
            assert (
                record.message
                == "The PC 127.0.0.1 tried to connect to TCP port 54996 on your PC without your permission."
            )
            assert record.keywords == "127.0.0.1,127.0.0.1,TCP port 54996"

        else:
            assert record.threat == "EICAR test file"
            assert record.ts == dt("2023-03-07T10:55:18Z")
            assert record.fkey == "{37E1F90E-471D-40D3-9FAA-37BE30C5B4AA}"
            assert (
                record.message
                == "Status Quarantined Scan type Custom  We found one or several threats on your PC. Threat name EICAR test file File C:\\Users\\admin\\Desktop\\eicar.com"  # noqa: E501
            )
            assert record.keywords == "Custom,EICAR test file,Quarantined"



from __future__ import annotations

import hashlib
from typing import TYPE_CHECKING

from dissect.target.plugins.os.windows import certlog
from dissect.target.target import Target
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem


def test_certlog_plugin(target_win: Target, fs_win: VirtualFilesystem) -> None:
    ca_edb = absolute_path("_data/plugins/os/windows/certlog/SEVENKINGDOMS-CA.edb")

    fs_win.map_file("Windows/System32/Certlog/SEVENKINGDOMS-CA.edb", ca_edb)

    target_win.add_plugin(certlog.CertLogPlugin)
    assert len(list(target_win.certlog())) == 142
    assert len(list(target_win.certlog.requests())) == 11
    assert len(list(target_win.certlog.request_attributes())) == 26
    assert len(list(target_win.certlog.crls())) == 2
    certificates = list(target_win.certlog.certificates())
    assert len(certificates) == 11
    assert len(list(target_win.certlog.certificate_extensions())) == 92
    assert certificates[0].serial_number == 23146941333149199441888068127529844838
    assert certificates[0].serial_number_hex == "1169f0517d9b598e4ba7af46e4674066"
    assert certificates[0].fingerprint.sha1 == hashlib.sha1(certificates[0].raw_certificate).hexdigest()
    assert certificates[0].fingerprint.sha1 == "061ec97dcef82d0e2d100b590e790a803d4573ae"


def test_certlog_plugin_direct() -> None:
    ca_edb = absolute_path("_data/plugins/os/windows/certlog/SEVENKINGDOMS-CA.edb")

    target = Target.open_direct([ca_edb])
    assert len(list(target.certlog())) == 142
    assert len(list(target.certlog.requests())) == 11
    assert len(list(target.certlog.request_attributes())) == 26
    assert len(list(target.certlog.crls())) == 2
    assert len(list(target.certlog.certificates())) == 11
    assert len(list(target.certlog.certificate_extensions())) == 92



from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING

from dissect.target.plugins.child.lima import LimaChildTargetPlugin

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_child_colima(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file_fh("/root/.lima/default/diffdisk", BytesIO())
    fs_unix.map_file_fh("/root/.lima/docker/diffdisk", BytesIO())
    fs_unix.map_file_fh("/home/user/.config/lima/ligma/diffdisk", BytesIO())

    target_unix_users.add_plugin(LimaChildTargetPlugin)

    children = [child for _, child in target_unix_users.list_children()]

    assert len(children) == 3

    assert children[0].type == "lima"
    assert children[0].name == "default"
    assert children[0].path == "/root/.lima/default/diffdisk"

    assert children[1].name == "docker"
    assert children[1].path == "/root/.lima/docker/diffdisk"

    assert children[2].path == "/home/user/.config/lima/ligma/diffdisk"



from __future__ import annotations

import json

import pytest

from dissect.target.tools.info import main as target_info
from tests._utils import absolute_path


@pytest.mark.parametrize(
    ("output_type", "options"),
    [("json", ["-j"]), ("json", ["-J", "-L", "tar"]), ("record", ["-r", "-s"])],
)
def test_target_info(
    capsys: pytest.CaptureFixture, monkeypatch: pytest.MonkeyPatch, output_type: str, options: list
) -> None:
    with monkeypatch.context() as m:
        m.setattr("sys.argv", ["target-info", *options, str(absolute_path("_data/tools/info/image.tar"))])

        target_info()
        stdout, stderr = capsys.readouterr()

        assert "error" not in stderr

        if output_type == "json":
            assert json.loads(stdout) == {
                "disks": [],
                "volumes": [],
                "children": [],
                "hostname": "ubuntu",
                "domain": None,
                "ips": ["1.2.3.4"],
                "os_family": "linux",
                "os_version": "Ubuntu 22.04.4 LTS (Jammy Jellyfish)",
                "architecture": None,
                "language": [],
                "timezone": "Europe/Amsterdam",
                "install_date": "2024-07-02 12:00:56+00:00",
                "last_activity": None,
                "mounts": [{"fs": "virtual", "path": "/"}],
            }

        elif output_type == "record":
            assert (
                stdout
                == "<target/info hostname='ubuntu' domain=None last_activity=None install_date=2024-07-02 12:00:56+00:00 ips=[net.ipaddress('1.2.3.4')] os_family='linux' os_version='Ubuntu 22.04.4 LTS (Jammy Jellyfish)' architecture=None language=[] timezone='Europe/Amsterdam' disks=[] volumes=[] mounts=[\"{'fs': 'virtual', 'path': '/'}\"] children=[]>\n"  # noqa: E501
            )

        else:
            raise ValueError("unknown output_type %s", output_type)


@pytest.mark.parametrize(
    ("output_type", "options"),
    [("json", ["-j"]), ("json", ["-J", "-L", "tar"]), ("record", ["-r", "-s"])],
)
def test_target_info_with_exception(
    capsys: pytest.CaptureFixture, monkeypatch: pytest.MonkeyPatch, output_type: str, options: list
) -> None:
    """Test target-info on a target that raises exceptions in some properties."""
    with monkeypatch.context() as m:
        m.setattr(
            "sys.argv",
            ["target-info", *options, str(absolute_path("_data/loaders/acquire/test-windows-fs-c-absolute.tar"))],
        )

        target_info()
        stdout, stderr = capsys.readouterr()

        assert "error" not in stderr

        if output_type == "json":
            assert json.loads(stdout) == {
                "disks": [],
                "volumes": [],
                "children": [],
                "hostname": None,
                "domain": None,
                "ips": None,
                "os_family": "windows",
                "os_version": None,
                "architecture": None,
                "language": None,
                "timezone": None,
                "install_date": None,
                "last_activity": None,
                "mounts": [{"fs": "virtual", "path": "c:"}, {"fs": "virtual", "path": "sysvol"}],
            }

        elif output_type == "record":
            assert (
                stdout
                == "<target/info hostname=None domain=None last_activity=None install_date=None ips=[] os_family='windows' os_version=None architecture=None language=[] timezone=None disks=[] volumes=[] mounts=[\"{'fs': 'virtual', 'path': 'c:'}\", \"{'fs': 'virtual', 'path': 'sysvol'}\"] children=[]>\n"  # noqa: E501
            )

        else:
            raise ValueError("unknown output_type %s", output_type)
